/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file dist_list.tpp
 * @brief Implementation of templete functions/classes.
 */
#ifndef __DIST_LIST_TPP__
#define __DIST_LIST_TPP__

#include <cassert>
#include <cstring>
#include "transporter.h"

namespace sketo {
  //------------------------------------------------------------------------
  // class dist_list

  template <typename A>
  dist_list<A>::dist_list()
    : data(NULL)
  {
  }

  template <typename A>
  dist_list<A>::dist_list(int size)
  {
    data = new _impl::dist_list_buffer<A>(size);
  }

  template <typename A>
  dist_list<A>::dist_list(int size, const A* seq_data)
  {
    data = new _impl::dist_list_buffer<A>(size);
    distribute(seq_data, size, 0);
    
  }

  template <typename A>
  dist_list<A>::dist_list(const dist_list<A> &that)
    : data(that.data)
  {
    if (data) data->increment_ref_count();
  }

  template <typename A>
  dist_list<A>::~dist_list()
  {
    if (data) {
      data->decrement_ref_count();
      if(data->ref_count == 0) delete data;
    }
  }

  template <typename A>
  dist_list<A> &dist_list<A>::operator=(const dist_list<A>& that)
  {
    // Do nothing if the input is the same as the output.
    if (this->data == that.data) return *this;

    if (data) {
      data->decrement_ref_count();
      if(data->ref_count == 0) delete data;
    }
    data = that.data;
    if (data) data->increment_ref_count();
    return *this;
  }

  template <typename A>
  A
  dist_list<A>::get(int index) const
  {
    A ret;
    int r, l, i;
    data->translate_global_index_to_local(index, &r, &l, &i);
    if (sketo::rank == r) {
      ret = data->bufs[l][i];
    }
    _impl::transporter::bcast(&ret, 1, r);
    return ret;
  }
  
  template <typename A>
  void
  dist_list<A>::set(int index, A val)
  {
    int r, l, i;
    data->translate_global_index_to_local(index, &r, &l, &i);
    if (sketo::rank == r) {
      data->bufs[l][i] = val;
    }
  }

  template <typename A>
  dist_list<A>
  dist_list<A>::clone() const
  {
    dist_list<A> ret(0);

    ret.data->bufs = data->bufs;
    ret.data->dirty = data->dirty;
    if(!data->dirty) {
      ret.data->sizes = data->sizes;
      ret.data->global_size = data->global_size;
    }

    return ret;
  }

  template <typename A>
  void
  dist_list<A>::gather(A* buffer) const
  {
    char nonnull = (buffer) ? 1 : 0;
    char* nonnulls = new char[sketo::procs];
    _impl::transporter::allgather(&nonnull, 1, nonnulls);

    int lb = 0;
    for (int l = 0; l < data->bufs.size(); l++) {
      for (int p = 0; p < sketo::procs; ++p) {
        if (p == sketo::rank) {
          int ls = data->bufs[l].size(); 
          if (nonnull) {
            for(int i = 0; i < ls; i++) buffer[lb+i] = data->bufs[l][i];
          }
          lb += ls;
          for (int pp = 0; pp < sketo::procs; ++pp) {
            if (pp == sketo::rank) continue;
            if (nonnulls[pp]) {
              _impl::transporter::send< int >(&ls, 1, pp, _impl::TAG_GATHER_SIZE);
              _impl::transporter::send< A >(static_cast<A*>(&(*data->bufs[l].begin())), ls, pp, _impl::TAG_GATHER);
            }
          }
        } else {
          if (nonnull) {
            int ls = 0;
            _impl::transporter::recv< int >(&ls, 1, p, _impl::TAG_GATHER_SIZE);
            _impl::transporter::recv< A >(buffer + lb, ls, p, _impl::TAG_GATHER);
            lb += ls;
          }
        }
      }
    }
  }
  template <typename A>
  void
  dist_list<A>::distribute(const A* buffer, int n, int org)
  {
    data->even_resize(n);

    if(sketo::rank == org) {
      for (int p = 0; p < sketo::procs; ++p) {
        int lb, ls;
        _impl::dist_list_buffer<A>::compute_local_array(n, sketo::procs, p, &lb, &ls);
        if(sketo::rank == p) {
          for(int i = 0; i < ls; i++) data->bufs[0][i] = buffer[lb+i];
        } else {
          _impl::transporter::send< A >(buffer + lb, ls, p, _impl::TAG_GATHER);
        }
      }
    } else {
      int lb, ls;
      _impl::dist_list_buffer<A>::compute_local_array(n, sketo::procs, sketo::rank, &lb, &ls);
      _impl::transporter::recv< A >(static_cast<A*>(&(*data->bufs[0].begin())), ls, org, _impl::TAG_GATHER);
    }
  }

  template <typename A>
  void
  dist_list<A>::print(std::ostream &stream) const
  {
    int n = get_global_size();
    A *buf = (sketo::rank == 0) ? new A[n] : NULL;
    gather(buf);
    if (sketo::rank == 0 ) {
      for (int i = 0; i < n; ++i) {
        stream << buf[i] << ' ';
      }
      stream << std::endl;
    }
    if(buf) delete [] buf;
  }


  template <typename A>
  void
  dist_list<A>::push_front(const A&a)
  {
    if(sketo::rank == 0) {
      data->bufs[0].insert(data->bufs[0].begin(), a);
    }
    if(!data->dirty) {
      data->sizes[0][0]++;
      data->global_size++;
    }
  }
  template <typename A>
  void
  dist_list<A>::push_back(const A&a)
  {
    if(sketo::rank == sketo::procs-1) {
      data->bufs.back().push_back(a);
    }
    if(!data->dirty) {
      data->sizes.back().back()++;
      data->global_size++;
    }
  }
  template <typename A>
  A
  dist_list<A>::pop_front()
  {
    A ret;
    int r, l, i;
    data->translate_global_index_to_local(0, &r, &l, &i);
    if (sketo::rank == r) {
      ret = data->bufs[l].front();
      data->bufs[l].erase(data->bufs[l].begin());
    }
    _impl::transporter::bcast(&ret, 1, r);
    if(!data->dirty) {
      data->sizes[l][r]--;
      data->global_size--;
    }
    return ret;
  }

  template <typename A>
  A
  dist_list<A>::pop_back()
  {
    A ret;
    int r, l, i;
    data->translate_global_index_to_local_rev(0, &r, &l, &i);
    if (sketo::rank == r) {
      ret = data->bufs[l].back();
      data->bufs[l].pop_back();
    }
    _impl::transporter::bcast(&ret, 1, r);
    if(!data->dirty) {
      data->sizes[l][r]--;
      data->global_size--;
    }
    return ret;
  }

  template <typename A>
  dist_list<A>
  dist_list<A>::rebalance(int n) const
  {
    std::vector<std::vector<int> > sizes = data->get_sizes();
    int nn = data->get_global_size();

    if(n < 0) n = nn;  // no removal of elements

    // now, n is the target size, and nn is the current size (= get_global_size()).
    // the target size has to be less than or equal to the current size
    assert ( n <= nn ) ; 
    
    // (nn - n) elements are discarded
    int d = nn - n;
    for(int l = data->bufs.size() - 1; l >= 0 && d > 0; l--) {
      for(int r = sketo::procs - 1; r >= 0 && d > 0; r--) {
        int dd = std::min(d, sizes[l][r]);
        sizes[l][r] -= dd;
        d -= dd;
      }
    }

    int tr = 0;

    dist_list<A> ret(n);
    _impl::dist_list_buffer<A>* _data = ret.data;

    int lb, ls;
    _impl::dist_list_buffer<A>::compute_local_array(n, sketo::procs, tr, &lb, &ls);
    int lb2 = 0;
    for(int l = 0; l < data->bufs.size(); l++) {
      for(int r = 0; r < sketo::procs; r++) {
        int sent = 0;
        while(sizes[l][r] > sent) {
          int ss = std::min(sizes[l][r] - sent, (ls + lb - lb2)); // send size
          if(r == sketo::rank) {
            if(r == tr) { // self copy
              std::copy(data->bufs[l].begin() + sent, data->bufs[l].begin() + sent + ss, _data->bufs[0].begin()+(lb2-lb));
            } else {  // send
              _impl::transporter::send< A >(static_cast<A*>(&(*(data->bufs[l].begin() + sent))), ss, tr, _impl::TAG_GATHER);
            }
          } else if (tr == sketo::rank) { // receive
            _impl::transporter::recv< A >(static_cast<A*>(&(*(_data->bufs[0].begin() + (lb2-lb)))), ss, r, _impl::TAG_GATHER);
          }
          lb2 += ss;
          if(ls + lb == lb2) {
            tr++;
            _impl::dist_list_buffer<A>::compute_local_array(n, sketo::procs, tr, &lb, &ls);
          }
          sent += ss;
        }
      }
    }
    return ret;
  }
  
  template <typename A>
  void
  dist_list<A>::rebalance()
  {
    *this = rebalance(-1);
  }

#ifndef __SKETO_NO_FUSION__
  template <typename A> template <typename B>
  dist_list<A>::dist_list(const B& that)
  {
    data = NULL;
    this->operator=(that);
  }
  template <typename A, bool overwriting = false>
  struct dist_list_copy_comp {
    _impl::dist_list_buffer<A> *data;
    dist_list_copy_comp(_impl::dist_list_buffer<A>*data) : data(data) { }
    A*buf;
    inline void init(int l, int ls) {
      if(!overwriting) {
        data->bufs.push_back(BUFFER<A>());
        data->bufs.back().resize(ls);
      }
      buf = &(data->bufs[l][0]);
    }
    inline void receive(int i, const A& a) {
      buf[i] = a;
    }
    inline void fin(int) { }
  };

  template <typename A, typename B, bool overwriting>
  struct dist_list_builder {
    static void build(_impl::dist_list_buffer<A> *data, const B &that) {
      dist_list_copy_comp<A, overwriting> comp(data);
      that.make_local_get_ready(); 
      if(overwriting) _impl::ls_adpt::dist_list_driver_ow(comp, that);
      else            _impl::ls_adpt::dist_list_driver(comp, that);
    }
  };
  template <typename A, typename OPLUS, typename AS, typename G, bool overwriting>
  struct dist_list_builder<typename G::result_type, _impl::ls_scanobj<OPLUS, AS, A, G>, overwriting> {
    static void build(_impl::dist_list_buffer<typename G::result_type> *data, const _impl::ls_scanobj<OPLUS, AS, A, G> &that) {
      that.template special_impl<overwriting>(data);
    }
  };
  template <typename A, typename OPLUS, typename AS, typename G,  bool overwriting>
  struct dist_list_builder<typename G::result_type, _impl::ls_scanrobj<OPLUS, AS, A, G>, overwriting> {
    static void build(_impl::dist_list_buffer<typename G::result_type> *data, const _impl::ls_scanrobj<OPLUS, AS, A, G> &that) {
      that.template special_impl<overwriting>(data);
    }
  };

  template <typename A>
  template <typename B>
  dist_list<A> &dist_list<A>::operator=(const B& that)
  {
    // overwriting judgement is not enough, because counting the actual uses of 'data' is difficult: count_ref is accurate only when the depth of 'that' is 1. 
    if(data && (data->ref_count - that.count_ref(data) == 1)) {
      if(!that.is_dirty() && !is_dirty()) {
        if (get_global_size() == that.get_global_size()) {
          // now, we can use overwriting.
          // however, if 'that' has its concrete list, then reuse it.
          if(that.has_concrete_list()) {
            //sketo::cout << "use concrete one!" << std::endl;
            *this = that.get_concrete_list();
            return *this;
          }

          dist_list_builder<A, B, true>::build(data, that);
	  
          data->dirty = false;
          data->sizes = that.get_sizes();
          data->global_size = that.get_global_size();
          return *this;
        }
      }
    }

    that.make_local_get_ready(); // make sure that 'that' is ready for copy.
    // if 'that' has been created its materialized list,
    // then just use (share) it!
    if(that.has_concrete_list()) {
      //sketo::cout << "use concrete one!" << std::endl;
      *this = that.get_concrete_list();
      return *this;
    }

    // allocate new buffer
    if (data) {
      data->decrement_ref_count();
      if(data->ref_count == 0) delete data;
    }
    data = new _impl::dist_list_buffer<A>(0);
    data->bufs.clear();

    dist_list_builder<A, B, false>::build(data, that);

    data->dirty = that.is_dirty();
    if(!data->dirty) {
      data->sizes = that.get_sizes();
      data->global_size = that.get_global_size();
    }
    //data->make_bufsp();
    return *this;
  }
#endif
  

  //------------------------------------------------------------------------
  // class dist_list_buffer
  template <typename A>
  _impl::dist_list_buffer<A>::dist_list_buffer(int gsize)
    : ref_count(1) //, bufsp(NULL)
  {
    even_resize(gsize);
  }
  
  template <typename A>
  _impl::dist_list_buffer<A>::~dist_list_buffer()
  {
    assert(ref_count == 0);
    //if(bufsp) delete bufsp;
    //if(buf) delete [] buf;
  }

  template <typename A>
  int
  _impl::dist_list_buffer<A>::get_global_size()
  {
    if(dirty) get_sizes();
    return global_size;
  }

  template <typename A>
  void
  _impl::dist_list_buffer<A>::even_resize(int n)
  {
    //buf = NULL;
    sizes.clear();
    sizes.push_back(std::vector<int>(sketo::procs));
    for(int r = 0; r < sketo::procs; r++) {
      int lb, ls;
      compute_local_array(n, sketo::procs, r, &lb, &ls);
      sizes[0][r] = ls;
    }
    dirty = false;  // currently, we know the correct sizes
    int lb, ls;
    compute_local_array(n, sketo::procs, sketo::rank, &lb, &ls);
    bufs.clear();
    bufs.push_back(BUFFER<A>());
    bufs.back().resize(ls);
    global_size = n;
    //make_bufsp();
  }
  template <typename A>
  void
  _impl::dist_list_buffer<A>::increment_ref_count()
  {
    assert(ref_count > 0);
    ++ref_count;
  }

  template <typename A>
  void
  _impl::dist_list_buffer<A>::decrement_ref_count()
  {
    assert(ref_count > 0);
    --ref_count;
    if (ref_count == 0) {
      // bufs.clear();
      // sizes.clear();
      // dirty = true;
      //     delete [] buf; buf = NULL;
    }
  }

  template <typename A>
  void
  _impl::dist_list_buffer<A>::compute_local_array(int global_size,
                                                  int procs,
                                                  int rank,
                                                  int* local_begin,
                                                  int* local_size)
  {
    const int div = global_size / procs;
    const int mod = global_size % procs;
    if (rank >  mod)    { *local_begin = div * rank + mod;  *local_size = div;     return; }
    if (rank == mod)    { *local_begin = (div + 1) * rank;  *local_size = div;     return; }
    /* (rank <  mod) */ { *local_begin = (div + 1) * rank;  *local_size = div + 1; return; }
  }

  template <typename A>
  int
  _impl::dist_list_buffer<A>::compute_rank_from_index(int global_size,
                                                      int procs,
                                                      int index)
  {
    const int div = global_size / procs;
    const int mod = global_size % procs;
    int p = 0;
    int begin = 0;
    int next = div + (p >= mod ? 0 : 1);
    for (; p < procs; ++p) {
      if (begin <= index && index < next) break;
      begin = next; next += div + (p + 1 >= mod ? 0 : 1);
    }
    return p;
  }

  template <typename A>
  void
  _impl::dist_list_buffer<A>::gather_sizes() 
  {
    sizes.clear();
    int nn = 0;
    for(int l = 0; l < bufs.size(); l++) {
      std::vector<int> s;
      sizes.push_back(s);
      sizes.back().resize(sketo::procs);
      int size = bufs[l].size();
      _impl::transporter::allgather(&size, 1, &(*sizes.back().begin()));
      for(int i = 0; i < sketo::procs; i++) nn += sizes.back()[i];
    }
    global_size = nn;
    dirty = false;
  }


  template <typename A>
  void 
  _impl::dist_list_buffer<A>::translate_global_index_to_local(int gi, int *r, int *l, int *i) const
  {
    if(dirty) {
      int* sizes = new int[sketo::procs];
      for(*l = 0; *l < bufs.size(); (*l)++) {
        int size = bufs[*l].size();
        _impl::transporter::allgather(&size, 1, sizes);
        for(*r = 0; *r < sketo::procs; (*r)++) {
          if(gi < sizes[*r]) {
            *i = gi; return;
          }
          gi -= sizes[*r];
        } 
      }
      delete [] sizes;
    } else {
      for(*l = 0; *l < bufs.size(); (*l)++) {
        for(*r = 0; *r < sketo::procs; (*r)++) {
          if(gi < sizes[*l][*r]) {
            *i = gi; return;
          }
          gi -= sizes[*l][*r];
        } 
      }
    }
  }

  template <typename A>
  void 
  _impl::dist_list_buffer<A>::translate_global_index_to_local_rev(int gi, int *r, int *l, int *i) const
  {
    if(dirty) {
      int* sizes = new int[sketo::procs];
      for(*l = bufs.size() - 1; *l >= 0; (*l)--) {
        int size = bufs[*l].size();
        _impl::transporter::allgather(&size, 1, sizes);
        for(*r = sketo::procs - 1; *r >= 0; (*r)--) {
          if(gi < sizes[*r]) {
            *i = gi; return;
          }
          gi -= sizes[*r];
        } 
      }
      delete [] sizes;
    } else {
      for(*l = bufs.size() - 1; *l >= 0; (*l)--) {
        for(*r = sketo::procs - 1; *r >= 0; (*r)--) {
          if(gi < sizes[*l][*r]) {
            *i = gi; return;
          }
          gi -= sizes[*l][*r];
        } 
      }
    }
  }

}

#endif // __DIST_LIST_TPP__
