/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file dist_matrix.h
 *
 * @brief The header for distributed matrices.
 */
#ifndef __DIST_MATRIX_H__
#define __DIST_MATRIX_H__

#include "transporter.h"

#include <utility>
#include "dist_matrix_predef.h"

namespace sketo {
/**
 * @brief A structure for matrix indices.
 */
  struct mindex {
    const int i;
    const int j;
    mindex(int iindex, int jindex) : i(iindex), j(jindex) {}
    mindex() : i(0), j(0) {}
    mindex & operator=(const mindex &that) { // dirty...
      *(int*)&i = that.i;
      *(int*)&j = that.j;
      return *this;
    }
    bool operator==(const mindex &that) const { 
      return (i == that.i) && (j == that.j);
    }
    bool operator!=(const mindex &that) const { 
      return !(*this == that);
    }
  private:
    friend class sketo::serialization::access;
    template<class Archive>
    void serialize(Archive&  ar, const unsigned int version) {
      ar & (*(int*)&i); // dangerous?
      ar & (*(int*)&j);
    }    
  };

  mindex mindex_zero(0, 0);

//typedef std::pair<int, int> mindex;

/**
 * @brief A structure for matrix sizes.
 */
//typedef std::pair<int, int> msize;
  struct msize {
    const int i;
    const int j;
    msize(int isize, int jsize) : i(isize), j(jsize) {}
    msize() : i(0), j(0) {}
    msize & operator=(const msize &that) { // dirty...
      *(int*)&i = that.i;
      *(int*)&j = that.j;
      return *this;
    }
    bool operator==(const msize &that) const { 
      return (i == that.i) && (j == that.j);
    }
    bool operator!=(const msize &that) const { 
      return !(*this == that);
    }
  private:
    friend class sketo::serialization::access;
    template<class Archive>
    void serialize(Archive&  ar, const unsigned int version) {
      ar & (*(int*)&i);
      ar & (*(int*)&j);
    }    
  };

/**
 * @brief Arithmetic of mindex and msize.
 */
  inline mindex operator+(const mindex index, const msize size)
  {
    return mindex(index.i + size.i, index.j + size.j);
  }

/**
 * @brief Arithmetic of mindex and msize.
 */
  inline mindex operator-(const mindex index, const msize size)
  {
    return mindex(index.i - size.i, index.j - size.j);
  }

/**
 * @brief Arithmetic of mindex and mindex.
 */
  inline mindex operator+(const mindex index, const mindex index2)
  {
    return mindex(index.i + index2.i, index.j + index2.j);
  }

/**
 * @brief Arithmetic of mindex and mindex.
 */
  inline mindex operator-(const mindex index, const mindex index2)
  {
    return mindex(index.i - index2.i, index.j - index2.j);
  }

/**
 * @brief Printout of mindex.
 */
  std::ostream &operator<<(std::ostream &os, const mindex index)
  {
    os << "(" << index.i << ", " << index.j << ")";
    return os;
  }

/**
 * @brief Printout of msize.
 */
  std::ostream &operator<<(std::ostream &os, const msize size)
  {
    os << "(" << size.i << ", " << size.j << ")";
    return os;
  }



/**
 * @brief A structure for pairs of matrix indices and sizes.
 */
//typedef std::pair<mindex, msize> mindexsize;
  struct mindexsize {
    const mindex index;
    const msize size;
    mindexsize(mindex _index, msize _size) : index(_index), size(_size) {}
    mindexsize() {}
    mindexsize & operator=(const mindexsize &that) { // dirty...
      *(mindex*)&index = that.index;
      *(msize*)&size = that.size;
      return *this;
    }
  private:
    friend class sketo::serialization::access;
    template<class Archive>
    void serialize(Archive&  ar, const unsigned int version) {
      ar & index;
      ar & size;
    }    
  };

  /**
   * @brief A class for distributed matrix (two-dimensional array, row-major-serialization).
   * 
   */
  template <typename A>
  class dist_matrix {
    friend class _impl::ms_adpt;
    
  public:
    typedef A element_type;

  private:
    _impl::dist_matrix_buffer<A> *data;

  public:
    dist_matrix();
    
    /** @brief A constructor that allocates distributed space */
    dist_matrix(msize size);

    /** @brief A constructor that allocates distributed space and copy from (row-major-serialized) data.
     * If org < 0, then each process copies a certain region from its own data; it is required that all processes have the same data.
     * Otherwise the data in process org is distributed among all processors.
     */
    dist_matrix(msize size, const A* seq_data, int org = -1);

    /** @brief A copy constructor
     *
     * Actually, this constructor does "shallow copy", that is, the two matrices (this and that)
     * share the actual buffer on memary.  Use clone() in addition if you want "deep copy".
     *
     * @param that the reference to another distributed matrix
     */
    dist_matrix(const dist_matrix<A> &that);

    ~dist_matrix();

    /** @brief Assignement
     *
     * Actually, this assignment does "shallow copy", that is, the two matrices (this and that)
     * share the actual buffer on memary.  Use clone() in addition if you want "deep copy".
     *
     * @param that the reference to another distributed matrix
     */
    dist_matrix<A>& operator=(const dist_matrix<A>& that);

  public:
    /** @brief Return the global size of the matrix */
    msize get_global_size() const;

    /** @brief Return the size of elements allocated to the process */
    msize get_local_size() const;
    
    // /** @brief Get the value of the specified global index.
    //  * @param iindex the global iindex
    //  * @param jindex the global jindex
    //  */
    // A get(int iindex, int jindex) const
    //   { return get(mindex(iindex, jindex)); }

    // /** @brief For skeleton implementation.
    //  * @param iindex the local iindex
    //  * @param jindex the local jindex 
    //  */
    // A local_get(int iindex, int jindex) const
    //   { return local_get(mindex(iindex, jindex)); }

    // /** @brief Set a value to the specifed global index.
    //  * @param iindex the global iindex.
    //  * @param jindex the global jindex.
    //  * @param val   value to be set.
    //  */
    // void set(int iindex, int jindex, A val)
    //   {set(mindex(iindex, jindex), val);}

    /** @brief Get the value of the specified global index.
     * @param index the global index
     */
    A get(mindex index) const;

    /** @brief For skeleton implementation.
     * @param index the local index 
     */
    A local_get(mindex index) const;

    /** @brief Set a value to the specifed global index.
     * @param index the global index.
     * @param val   value to be set.
     */
    void set(mindex index, A val);

    /** @brief Set a value to the specifed local index.
     * @param index the local index.
     * @param val   value to be set.
     */
    void local_set(mindex index, A val);

    /** @brief Create a (deep) copy of the matrix. */
    dist_matrix<A> clone() const;
        
    /** @brief Gather the distributed values to the (sequential) array.
     *
     * This function gathers the distributed values to the array specified by the input.
     * The user should prepare enough space (i.e., get_global_size()).
     *
     * For reasons of performance, user can specify the buffer as NULL on some processes,
     * then this function does not gather the values to the process.
     * For example, this is effective when the gathered values are only used in outputting them
     * from the root process.
     *
     * @param buffer The pointer to the array to which the values are gathered, or NULL.
     */
    void gather(A* buffer) const;

    
      
#ifndef __SKETO_NO_FUSION__
    // The following are used when fusion transformation is activated.
  public:
    template <typename B>
    dist_matrix(const B& that);
    
    template <typename B>
    dist_matrix<A>& operator=(const B& that);

    template <typename T>
    void check_overwrite(_impl::dist_matrix_buffer<T>* target,
                         int *refcount, bool *, bool *, bool *scan_shift) const
      {
        if ((void*)data == (void*)target) (*refcount)++;
        // *no_deps = true; *child_no_deps = ---;
        *scan_shift = false;
      }
      
#endif // __SKETO_NO_FUSION__

  private:
    
    /** @brief Shallow copy of a rectangle region from a matrix buffer to a matrix buffer.
     *
     * @param buf_from   source row-major buffer
     * @param size_from  2D size of the source
     * @param index_from the start 2D in the source
     * @param buf_to     destination row-major buffer
     * @param size_to    2D size of the destination
     * @param index_to   the start 2D in the destination
     * @param size       the 2D size of the rectangle region to be copied
     */
    void
    copy_buffer(const A* buf_from, msize size_from, mindex index_from, A* buf_to, msize size_to, mindex index_to, msize size) const;
    
    /** @brief Creates a region_serializer for serialization of a rectangle region in a matrix buffer to a linear buffer.

     * If the size of the linear buffer is not sufficient,
     * this function automatically allocate a new buffer.
     *
     * @param buf_from   source row-major buffer
     * @param size_from  2D size of the source
     * @param index_from the start 2D in the source
     * @param size       the 2D size of the rectangle region to be copied
     */
    sketo::serialization::region_serializer<A>
    make_serializer_for_serialize_buffer(const A* buf_from, msize size_from, mindex index_from, msize size) const;

    /** @brief Creates a region_serializer for deserialization of a linear buffer to a rectangle region in a matrix buffer.
     *
     * @param buf_to     destination row-major buffer
     * @param size_to    2D size of the destination
     * @param index_to   the start 2D in the destination
     * @param size       the 2D size of the rectangle region to be copied
     */
    sketo::serialization::region_serializer<A>
    make_serializer_for_deserialize_buffer(A* buf_to, msize size_to, mindex index_to, msize size) const;

    /** @brief This function ditributes the given array among processors.
     *
     * @param size   The size of the array.
     * @param buffer The pointer to the given elements to be ditributed.
     * @param org    The rank of the process that has the elements .
     */
    void
    distribute(msize size, const A* buffer, int org = 0);
  };

  namespace _impl {

    /**
     * @brief The main data structure for maintaining distributed matrix.
     */
    template <typename A>
    class dist_matrix_buffer {
      friend class dist_matrix<A>;
      friend class _impl::ms_adpt;

    private:
      int ref_count;

      msize global_size;
      A* buf;
      
      mindex local_begin;
      msize local_size;

    public:
      dist_matrix_buffer(msize gsize);
      ~dist_matrix_buffer();

      msize get_global_size() const { return global_size; }
      msize get_local_size() const { return local_size; }
      mindex get_local_begin() const { return local_begin; }

      /** @brief Increment the reference count */
      void increment_ref_count();

      /** @brief Decrement the reference count and free the allocated memory when the count becomes zero */
      void decrement_ref_count();

    public:      // utility functions
      /**
       * @brief Compute the first local index and the local size in a direction of a distributed matrix.
       * @param global_size  global size of the dimension of the matrix
       * @param procs        the number of processes
       * @param rank         the rank of process
       * @param local_begin  first local index in the direction will be stored
       * @param local_size   local size in the direction will be stored
       */
      static void
      compute_local_array_stab(int global_size, int procs, int rank,
                               int* local_begin, int* local_size);

      /**
       * @brief Compute the first local index and the local size of a distributed matrix.
       * @param global_size  global size of the matrix
       * @param procs        the number of processes
       * @param rank         the rank of process
       */
      static mindexsize
      compute_local_array(msize global_size, int procs, int rank);

      /**
       * @brief Compute the first local index and the local size of a distributed matrix.
       * @param global_size  global size of the matrix
       * @param ps           the 2D process arrangement
       * @param rs           the 2D rank of process
       */
      static mindexsize
      compute_local_array(msize global_size, msize ps, mindex rs);
      
      /**
       * @brief Compute the process arrangement.
       * @param global_size  global size of the matrix
       * @param procs        the number of processes
       */
      static msize
      compute_process_arrangement(msize global_size, int procs);


      /**
       * @brief Compute a multi-dimensional rank.
       * @param pa     the process arrangement
       * @param rank   the rank of process
       */
      static mindex
      compute_ranks(msize pa, int rank);

      /**
       * @brief Compute a rank from a multi-dimensional rank.
       * @param pa     the process arrangement
       * @param rs     the multi-dimensional rank of process
       */
      static int
      compute_rank(msize pa, mindex rs);
      
      /**
       * @brief Compute the rank of process that has the value of specified index. The first component of the result is the 2D rank of the process, and the second is the pair of the local index on the process and the flag showing that the given index is folded.
       * @param global_size  global size of the matrix
       * @param procs        the number of processes
       * @param index        the index to be checked.
       * @param pa           the process arrangement
       */
      static std::pair<mindex, std::pair<mindex, bool> >
      compute_rank_from_index(msize global_size, int procs, mindex index, msize pa);

      /**
       * @brief Compute the rank of process that has the value of specified index in a direction. The first component of the result is the rank of the process, and the second ist he pair of the local index  in the direction on the process and the flag showing that the given index is folded.

       * @param global_size  global size in the direction of the matrix
       * @param procs        the number of processes in the direction
       * @param index        the index to be checked in the direction
       */
      static std::pair<int, std::pair<int, bool> >
      compute_rank_from_index_stab(int global_size, int procs, int index);
    };

  }

}

#include "dist_matrix.tpp"

#endif // __DIST_MATRIX_H__
