/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file dist_matrix.tpp
 * @brief Implementation of templete functions/classes.
 */
#ifndef __DIST_MATRIX_TPP__
#define __DIST_MATRIX_TPP__

#include <cassert>
#include <cstring>

//
// TODO: Separate direct asscesses (marked with /*****/) from dist_matrix.
//       We have mant choices to serialize a matrix: row/col/Z/N/...
//
// Now, we compute local array size of a node by compute_local_array, 
// in which the size is determined from procs, rank, and the global size. 
// Thus, we cannot do block-rotations in which the size needs to be determined
// from procs, rank, the global size, and the number of rotations.
// E.g., when we distribute 1x3 matrix to 2 nodes, node 1 has 1x2 and
// node 2 has 1x1. Then, after a rotation, they have 1x1 and 1x2, respectively.
// The size after the rotation cannot be determined without the information
// that there was a rotation.
//
//


namespace sketo {
  //------------------------------------------------------------------------
  // class dist_matrix

  template <typename A>
  dist_matrix<A>::dist_matrix()
    : data(NULL)
  {
  }

  template <typename A>
  dist_matrix<A>::dist_matrix(msize size)
  {
    data = new _impl::dist_matrix_buffer<A>(size);
  }

  template <typename A>
  dist_matrix<A>::dist_matrix(msize size, const A* seq_data, int org)
  {
    if(org < 0) {
      data = new _impl::dist_matrix_buffer<A>(size);
      mindexsize mis = _impl::dist_matrix_buffer<A>::compute_local_array(size, sketo::procs, sketo::rank);
      copy_buffer(seq_data, size, mis.index, data->buf, mis.size, mindex_zero, mis.size);
    } else {
      _impl::transporter::bcast(&size, 1, org);
      data = new _impl::dist_matrix_buffer<A>(size);
      distribute(size, seq_data, org);
    }
  }
  
  template <typename A>
  dist_matrix<A>::dist_matrix(const dist_matrix<A> &that)
    : data(that.data)
  {
    if (data) data->increment_ref_count();
  }

  template <typename A>
  dist_matrix<A>::~dist_matrix()
  {
    if (data) {
      data->decrement_ref_count();
      if(data->ref_count == 0) delete data;
    }
  }

  template <typename A>
  dist_matrix<A> &dist_matrix<A>::operator=(const dist_matrix<A>& that)
  {
    // Do nothing if the input is the same as the output.
    if (this->data == that.data) return *this;

    if (data) {
      data->decrement_ref_count();
      if(data->ref_count == 0) delete data;
    }
    data = that.data;
    if (data) data->increment_ref_count();
    return *this;
  }
  
  template <typename A>
  inline msize
  dist_matrix<A>::get_global_size() const
  {
    assert(data);
    return data->get_global_size();
  }

  template <typename A>
  inline msize
  dist_matrix<A>::get_local_size() const
  {
    assert(data);
    return data->get_local_size();
  }

  template <typename A>
  A
  dist_matrix<A>::get(mindex index) const
  {
    A ret;
    msize pa = _impl::dist_matrix_buffer<A>::compute_process_arrangement(get_global_size(), procs);
    mindex rs = _impl::dist_matrix_buffer<A>::compute_rank_from_index(get_global_size(), sketo::procs, index, pa).first;
    int target_p = _impl::dist_matrix_buffer<A>::compute_rank(pa, rs);
    if (sketo::rank == target_p) {
      mindex i = data->get_local_begin();
      ret = local_get(index - i);
    }
    _impl::transporter::bcast<A>(&ret, 1, target_p);
    return ret;
  }

  template <typename A>
  inline A
  dist_matrix<A>::local_get(mindex index) const
  {
    msize s = data->get_local_size();
    return data->buf[index.i * s.j + index.j]; /*****/
  }
  
  template <typename A>
  void
  dist_matrix<A>::set(mindex index, A val) 
  {
    msize pa = _impl::dist_matrix_buffer<A>::compute_process_arrangement(get_global_size(), procs);
    mindex rs = _impl::dist_matrix_buffer<A>::compute_rank_from_index(data->global_size, sketo::procs, index, pa).first;
    int target_p = _impl::dist_matrix_buffer<A>::compute_rank(pa, rs);
    if (sketo::rank == target_p) {
      mindex i = data->get_local_begin();
      local_set(index - i, val);
    }
  }

  template <typename A>
  void
  dist_matrix<A>::local_set(mindex index, A val) 
  {
    msize s = data->get_local_size();
    data->buf[index.i * s.j + index.j] = val; /*****/
  }


  template <typename A>
  dist_matrix<A>
  dist_matrix<A>::clone() const
  {
    dist_matrix<A> ret(get_global_size());
    A* const ret_data = ret.data->buf;
    msize s = get_local_size();
    const int nj = s.j;               /*****/
    const int ni = s.i;               /*****/
    for (int i = 0; i < ni; ++i) {
      for (int j = 0; j < nj; ++j) {
        ret_data[i * nj + j] = local_get(mindex(i, j));
      }
    }
    return ret;
  }

  template <typename A>
  void
  dist_matrix<A>::copy_buffer(const A* buf_from, msize size_from, mindex index_from, A* buf_to, msize size_to, mindex index_to, msize size) const
  {
    for(int i = 0; i < size.i; i++) { /*****/
      const A* jt = buf_from + (index_from.i + i) * size_from.j + index_from.j;
      A* it = buf_to + (index_to.i + i) * size_to.j + index_to.j;
      const A* const je = jt + size.j;
      for (; jt != je; ++it, ++jt) {
        *it = *jt;
      }
    }
  }
  
  template <typename A>
  sketo::serialization::region_serializer<A>
  dist_matrix<A>::
  make_serializer_for_serialize_buffer(const A* buf_from, msize size_from, mindex index_from, msize size) const
  {    
    sketo::serialization::region_serializer<A> rs(buf_from,
                                                  index_from.i * size_from.j + index_from.j,
                                                  size.j,
                                                  size_from.j - size.j,
                                                  size.i);
    return rs;
  }
  
  template <typename A>
  sketo::serialization::region_serializer<A>
  dist_matrix<A>::
  make_serializer_for_deserialize_buffer(A* buf_to, msize size_to, mindex index_to, msize size) const
  {
    sketo::serialization::region_serializer<A> rs(buf_to,
                                                  index_to.i * size_to.j + index_to.j,
                                                  size.j,
                                                  size_to.j - size.j,
                                                  size.i);
    return rs;
  }

  template <typename A>
  void
  dist_matrix<A>::gather(A* buffer) const
  {
    char nonnull = (buffer) ? 1 : 0;
    char* nonnulls = new char[sketo::procs];
    _impl::transporter::allgather(&nonnull, 1, nonnulls);

    mindexsize *miss = new mindexsize[sketo::procs];
    int maxsize = 0;
    for (int p = 0; p < sketo::procs; ++p) {
      miss[p] = _impl::dist_matrix_buffer<A>::compute_local_array(get_global_size(), sketo::procs, p);
      if(miss[p].size.j * miss[p].size.i > maxsize) {
        maxsize = miss[p].size.j * miss[p].size.i;
      }
    }    

    for (int p = 0; p < sketo::procs; ++p) {
      mindexsize mis = miss[p]; //= _impl::dist_matrix_buffer<A>::compute_local_array(get_global_size(), sketo::procs, p);
      
      if (p == sketo::rank) {
        if (nonnull) {
          copy_buffer(data->buf, mis.size, mindex_zero, buffer, get_global_size(), mis.index, mis.size);
        }
        for (int pp = 0; pp < sketo::procs; ++pp) {
          if (pp == sketo::rank) continue;
          if (nonnulls[pp]) {
            sketo::serialization::region_serializer<A> rs = make_serializer_for_serialize_buffer(static_cast<A*>(data->buf), mis.size, mindex_zero, mis.size);
            _impl::transporter::send< sketo::serialization::region_serializer<A> >(&rs, 1, pp, _impl::TAG_M_GATHER);
          }
        }
      } else {
        if (nonnull) {
          sketo::serialization::region_serializer<A> rs =  make_serializer_for_serialize_buffer(buffer, get_global_size(), mis.index, mis.size);
          _impl::transporter::recv< sketo::serialization::region_serializer<A> >(&rs, 1, p, _impl::TAG_M_GATHER);
        }
      }
    }
  }

  template <typename A>
  void
  dist_matrix<A>::distribute(msize size, const A* buffer, int org)
  {
    mindexsize *miss = new mindexsize[sketo::procs];
    for (int p = 0; p < sketo::procs; ++p) {
      miss[p] = _impl::dist_matrix_buffer<A>::compute_local_array(get_global_size(), sketo::procs, p);
    }  
    
    if(sketo::rank == org) {
      for (int p = 0; p < sketo::procs; ++p) {
        mindexsize mis = miss[p];
        if(sketo::rank == p) {
          copy_buffer(buffer, get_global_size(), mis.index, data->buf, mis.size, mindex_zero, mis.size);
        } else {
          sketo::serialization::region_serializer<A> rs =  make_serializer_for_serialize_buffer(buffer, get_global_size(), mis.index, mis.size);
          _impl::transporter::send< sketo::serialization::region_serializer<A> >(&rs, 1, p, _impl::TAG_M_GATHER);
        }
      }
    } else {
      mindexsize mis = miss[sketo::rank];
      sketo::serialization::region_serializer<A> rs =  make_serializer_for_deserialize_buffer(static_cast<A*>(data->buf), mis.size, mindex_zero, mis.size);
      _impl::transporter::recv< sketo::serialization::region_serializer<A> >(&rs, 1, org, _impl::TAG_M_GATHER);
    }
  }

#ifndef __SKETO_NO_FUSION__
  template <typename A> template <typename B>
  dist_matrix<A>::dist_matrix(const B& that)
  {
    data = NULL;
    this->operator=(that);
  }
  
  template <typename A>
  template <typename B>
  dist_matrix<A> &dist_matrix<A>::operator=(const B& that)
  {
    // Condition for overwriting:
    //   - The sizes of matrices are the same
    //   - Reference count is 1
    //   - No shift operations on the data
    int refcount = 0;
    bool no_deps = true;
    bool child_no_deps = true;
    bool scan_shift = false;
    that.check_overwrite(data, &refcount, &no_deps, &child_no_deps, &scan_shift);
    
    if ( data
         && this->get_global_size() == that.get_global_size()
         && data->ref_count - refcount == 1
         && ((scan_shift && child_no_deps)
             || (!scan_shift && no_deps))) {
      // do nothing.
    } else {
      // allocate new buffer
      if (data) {
         data->decrement_ref_count();
         if(data->ref_count == 0) delete data;
      }
      data = new _impl::dist_matrix_buffer<A>(that.get_global_size());
    }

    if (scan_shift && child_no_deps) {
      return *(that.ow_impl(this));
    } else {

      A* const this_data = this->data->buf;
      msize s = that.get_local_size();
      const int nj = s.j;               /*****/
      const int ni = s.i;               /*****/
      for (int i = 0; i < ni; ++i) {
        for (int j = 0; j < nj; ++j) {
          this_data[i * nj + j] = that.local_get(mindex(i, j));
        }
      }
      return *this;
    }
  }
#endif
  
  
  //------------------------------------------------------------------------
  // class dist_matrix_buffer
  template <typename A>
  _impl::dist_matrix_buffer<A>::dist_matrix_buffer(msize gsize)
    : ref_count(1), global_size(gsize)
  {
    mindexsize mis = compute_local_array(gsize, sketo::procs, sketo::rank);
    local_begin = mis.index;
    local_size = mis.size;
    buf = new A[local_size.j * local_size.i];
  }
  
  template <typename A>
  _impl::dist_matrix_buffer<A>::~dist_matrix_buffer()
  {
    assert(ref_count == 0);
  }

  template <typename A>
  void
  _impl::dist_matrix_buffer<A>::increment_ref_count()
  {
    assert(ref_count > 0);
    ++ref_count;
  }

  template <typename A>
  void
  _impl::dist_matrix_buffer<A>::decrement_ref_count()
  {
    assert(ref_count > 0);
    --ref_count;
    if (ref_count == 0) {
      delete [] buf; buf = NULL;
    }
  }

  template <typename A>
  void
  _impl::dist_matrix_buffer<A>::compute_local_array_stab(int global_size,
                                                         int procs,
                                                         int rank,
                                                         int *local_begin,
                                                         int *local_size)
  {
    const int div = global_size / procs;
    const int mod = global_size % procs;
    if (rank >  mod)    {*local_begin = div * rank + mod;  *local_size = div;     return; }
    if (rank == mod)    { *local_begin = (div + 1) * rank;  *local_size = div;     return; }
    /* (rank <  mod) */ { *local_begin = (div + 1) * rank;  *local_size = div + 1; return; }
  }

  /* this should be changable for matrices. such as row-div, col-div, sq-div..*/
  template <typename A>
  mindexsize
  _impl::dist_matrix_buffer<A>::compute_local_array(msize global_size,
                                                    msize pa,
                                                    mindex rs)
  {
    
    int index_i = 0, index_j = 0, size_i = 0, size_j = 0;
    const int iprocs = pa.i;
    const int jprocs = pa.j;
    const int irank = rs.i;
    const int jrank = rs.j;

    compute_local_array_stab(global_size.i, iprocs, irank, &index_i, &size_i);
    compute_local_array_stab(global_size.j, jprocs, jrank, &index_j, &size_j);

    //std::cout << "rank " << rank << " has a region of " << msize(size_i, size_j) << " at " << mindex(index_i, index_j) << std::endl;

    return mindexsize(mindex(index_i, index_j), msize(size_i, size_j));
  }
  template <typename A>
  mindexsize
  _impl::dist_matrix_buffer<A>::compute_local_array(msize global_size,
                                                    int procs,
                                                    int rank)
  {
    msize pa = compute_process_arrangement(global_size, procs);

    mindex rs = compute_ranks(pa, rank);
    return compute_local_array(global_size, pa, rs);
  }

  template <typename A>
  mindex
  _impl::dist_matrix_buffer<A>::compute_ranks(msize pa,
                                              int rank)
  {
    const int irank = rank / pa.j;
    const int jrank = rank % pa.j;

    return mindex(irank, jrank);    
  }

  template <typename A>
  int
  _impl::dist_matrix_buffer<A>::compute_rank(msize pa,
                                             mindex rs)
  {
    return rs.i * pa.j + rs.j;    
  }

  template <typename A>
  msize
  _impl::dist_matrix_buffer<A>::compute_process_arrangement(msize global_size,
                                                            int procs)
  {
    // fat-rectangle division....
    // TODO: The shape of global_size should be respected here.
    //       E.g., global_size = (1, n), then we'd like to use div of (1, procs)
    int d = 1;
    for(int dd = 1; procs / dd >= dd; dd++) {
      if (procs % dd == 0) {
        d = dd;
      }
    }
    const int iprocs = d;
    const int jprocs = procs / iprocs;
    return msize(iprocs, jprocs);
  }

  
  template <typename A>
  std::pair<mindex, std::pair<mindex, bool> >
  _impl::dist_matrix_buffer<A>::compute_rank_from_index(msize global_size,
                                                        int procs,
                                                        mindex index,
                                                        msize pa)
  {
    const int iprocs = pa.i;
    const int jprocs = pa.j;

    std::pair<int, std::pair<int, bool> > irankindex = compute_rank_from_index_stab(global_size.i, iprocs, index.i);
    std::pair<int, std::pair<int, bool> > jrankindex = compute_rank_from_index_stab(global_size.j, jprocs, index.j);
    //std::cout << "owner of " << index << " is " << mindex(irank, jrank) << std::endl;
    
    return std::pair<mindex, std::pair<mindex, bool> >(mindex(irankindex.first, jrankindex.first),
                                                       std::pair<mindex, bool>(mindex(irankindex.second.first, jrankindex.second.first), irankindex.second.second || jrankindex.second.second));
  }
  template <typename A>
  std::pair<int, std::pair<int, bool> >
  _impl::dist_matrix_buffer<A>::compute_rank_from_index_stab(int global_size,
                                                             int procs,
                                                             int index)
  {
    int nindex = ((index % global_size) + global_size) % global_size; // normalize
    bool flag = nindex != index;
    index = nindex;
    const int div = global_size / procs;
    const int mod = global_size % procs;
    int p = 0;
    int begin = 0;
    int next = div + (p >= mod ? 0 : 1);
    for (; p < procs; ++p) {
      if (begin <= index && index < next) break;
      begin = next; next += div + (p + 1 >= mod ? 0 : 1);
    }
    return std::pair<int, std::pair<int, bool> >(p, std::pair<int, bool>(index-begin, flag));
  }

}

#endif // __DIST_MATRIX_TPP__
