/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file list_skeletons_noopt.h
 * @brief An implmentation of list skeletons without fusion optimization.
 *        These skeletons are for debugging.
 */

#ifndef __LIST_SKELETONS_NOOPT_H__
#define __LIST_SKELETONS_NOOPT_H__

namespace sketo {

  namespace list_skeletons {
    /**
     * generate :: int -> (int -> a) -> [a]
     */
    template <typename F>
    dist_list<typename result_of<F(int)>::type>
    generate(int global_size,
	     const F& f);

    /**
     * map :: (a -> b) -> [a] -> [b]
     */
    template <typename F, typename A>
    dist_list<typename result_of<F(A)>::type>
    map(F f,
	const sketo::dist_list<A>& as);

    /**
     * map_with_index :: (int -> a -> b) -> [a] -> [b]
     */
    template <typename F, typename A>
    dist_list<typename result_of<F(int, A)>::type>
    map_with_index(const F& f,
		   const dist_list<A>& as);

    /**
     * zip :: [a] -> [b] -> [(a, b)]
     */
    template <typename A, typename B>
    dist_list<std::pair<A, B> >
    zip(const dist_list<A>& as,
	const dist_list<B>& bs);

    /**
     * zipwith :: (a -> b -> c) -> [a] -> [b] -> [c]
     */
    template <typename F, typename A, typename B>
    dist_list<typename result_of<F(A, B)>::type>
    zipwith(const F& f,
	    const dist_list<A>& as,
	    const dist_list<B>& bs);

    /**
     * concatmap :: (a -> [b]) -> [a] -> [b]
     */
    template <typename F, typename A>
    dist_list<typename result_of<F(A)>::type>
    concatmap(F f,
	      const sketo::dist_list<A>& as);

    /**
     * filter :: (a -> bool) -> [a] -> [a]
     */
    template <typename P, typename A>
    dist_list<A>
    filter(P p,
	   const sketo::dist_list<A>& as);


    /**
     * reduce :: (a -> a -> a) -> a -> [a] -> [a]
     */
    template <typename OPLUS, typename A>
    A
    reduce(const OPLUS &oplus,
	   const A &identity,
	   const dist_list<A>& as);

    /**
     * scan :: (a -> a -> a) -> a -> (a, [a]) -> ([a], a)
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    scan(const OPLUS &oplus,
	 const A &identity,
	 A e,
	 const dist_list<A>& as,
	 A *last_val = NULL);

    /**
     * scanr :: (a -> a -> a) -> a -> ([a], a) -> (a, [a])
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    scanr(const OPLUS &oplus,
	  const A &identity,
	  A e,
	  const dist_list<A>& as,
	  A *first_val = NULL);

    /**
     * postscan :: (a -> a -> a) -> a -> [a] -> [a]
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    postscan(const OPLUS &oplus,
	     const A &identity,
	     const dist_list<A>& as);

    /**
     * postscanr :: (a -> a -> a) -> a -> [a] -> [a]
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    postscanr(const OPLUS &oplus,
	      const A &identity,
	      const dist_list<A>& as);

    /**
     * gscanl :: (b -> a -> b) -> b -> (a -> c) -> (c -> c -> c) -> (b -> c -> b) -> c -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
    dist_list<B>
    gscanl(const F &f,
	   const B &c,
	   const G &g,
	   const OMINUS &ominus,
	   const OPLUS &oplus,
	   const typename OPLUS::result_type &identity,
	   const dist_list<A>& as);

    /**
     * gscanr :: (b -> a -> b) -> b -> (a -> c) -> (c -> c -> c) -> (b -> c -> b) -> c -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
    dist_list<B>
    gscanr(const F &f,
	   const B &c,
	   const G &g,
	   const OMINUS &ominus,
	   const OPLUS &oplus,
	   const typename OPLUS::result_type &identity,
	   const dist_list<A>& as);
    
    /**
     * shiftl :: ([a], a) -> (a, [a])
     */
    template <typename A>
    dist_list<A>
    shiftl(const A &rightmost, 
	   const dist_list<A> &as,
	   A *leftmost = NULL);

    /**
     * shiftr :: (a, [a]) -> ([a], a)
     */
    template <typename A>
    dist_list<A>
    shiftr(const A &leftmost,
	   const dist_list<A> &as,
	   A *rightmost = NULL);
    
    /**
     * append :: [a] -> [a] -> [a]
     */
    template <typename A>
    dist_list<A>
    append(const dist_list<A> &as,
	   const dist_list<A> &bs);

    /**
     * reduce :: (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename A>
    A
    reduce(const OPLUS &oplus,
	   const dist_list<A>& as);

    /**
     * scan :: (a -> a -> a) -> (a, [a]) -> ([a], a)
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    scan(const OPLUS &oplus,
	 A e,
	 const dist_list<A>& as,
	 A *last_val = NULL);

    /**
     * scanr :: (a -> a -> a) -> ([a], a) -> (a, [a])
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    scanr(const OPLUS &oplus,
	  A e,
	  const dist_list<A>& as,
	  A *first_val = NULL);

    /**
     * postscan :: (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    postscan(const OPLUS &oplus,
	     const dist_list<A>& as);

    /**
     * postscanr :: (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename A>
    dist_list<A>
    postscanr(const OPLUS &oplus,
	      const dist_list<A>& as);

    /**
     * gscanl :: (b -> a -> b) -> b -> (a -> c) -> (c -> c -> c) -> (b -> c -> b) -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
    dist_list<B>
    gscanl(const F &f,
	   const B &c,
	   const G &g,
	   const OMINUS &ominus,
	   const OPLUS &oplus,
	   const dist_list<A>& as);

    /**
     * gscanr :: (b -> a -> b) -> b -> (a -> c) -> (c -> c -> c) -> (b -> c -> b) -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
    dist_list<B>
    gscanr(const F &f,
	   const B &c,
	   const G &g,
	   const OMINUS &ominus,
	   const OPLUS &oplus,
	   const dist_list<A>& as);

  }    
}

#include "list_skeletons_noopt.tpp"

#endif //__LIST_SKELETONS_WITHOUT_FUSION_H__
