/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file matrix_skeletons_noopt.tpp
 * @brief Template implementation of matrix skeletons without fusion optimization.
 */
#ifndef __MATRIX_SKELETONS_NOOPT_TPP__
#define __MATRIX_SKELETONS_NOOPT_TPP__
#include <cassert>

#include "matrix_skeletons_common.tpp"

namespace sketo {
  //------------------------------------------------------------------------
  // generate
  template <typename F>
  dist_matrix<typename result_of<F(mindex)>::type>
  matrix_skeletons::generate(msize global_size, const F& f)
  {
    return _impl::ms_adpt::generate<F, typename result_of<F(int)>::type>(global_size, f);
  }

  template <typename F, typename A>
  dist_matrix<A>
  _impl::ms_adpt::generate(msize global_size,
			   const F& f)
  {
    dist_matrix<A> ret(global_size);

    A* const ret_it_end = ret.data->buf + ret.data->local_size.i * ret.data->local_size.j;
    mindex index = ret.data->local_begin;
    int i = index.i, j = index.j;
    const int je = j + ret.data->local_size.j;

    A* ret_it = ret.data->buf;
    
    for (;ret_it != ret_it_end; ++j, ++ret_it) {
      if( j == je ) {
        j = index.j;
        i++;
      }
      *ret_it = f(mindex(i, j));
    }
    
    return ret;
  }

  //------------------------------------------------------------------------
  // map
  template <typename F, typename A>
  dist_matrix<typename result_of<F(A)>::type>
  matrix_skeletons::map(F f,
		      const dist_matrix<A>& as)
  {
    return _impl::ms_adpt::map<F, A, typename result_of<F(A)>::type>(f, as);
  }

  template <typename F, typename A, typename B>
  dist_matrix<B>
  _impl::ms_adpt::map(const F& f,
		      const dist_matrix<A>& as)
  {
    dist_matrix<B> ret(as.get_global_size());

    B* const ret_buf = ret.data->buf;
    const int nj = as.get_local_size().j;
    const int ni = as.get_local_size().i;
    for (int i = 0; i < ni; ++i) {
      for (int j = 0; j < nj; ++j) {
        ret_buf[i * nj + j] = f(as.local_get(mindex(i, j)));
      }
    }
    
    return ret;
  }

  //------------------------------------------------------------------------
  // map_with_index
  template <typename F, typename A>
  dist_matrix<typename result_of<F(mindex, A)>::type>
  matrix_skeletons::map_with_index(const F& f,
				 const dist_matrix<A>& as)
  {
    return _impl::ms_adpt::map_with_index<F, A, typename result_of<F(mindex, A)>::type>(f, as);
  }

  template <typename F, typename A, typename B>
  dist_matrix<B>
  _impl::ms_adpt::map_with_index(const F& f,
				 const dist_matrix<A>& as)
  {
    dist_matrix<B> ret(as.get_global_size());

    B* const ret_buf = ret.data->buf;

    const mindex index = ret.data->local_begin;
    const int nj = as.get_local_size().j;
    const int ni = as.get_local_size().i;
    for (int i = 0; i < ni; ++i) {
      for (int j = 0; j < nj; ++j) {
        ret_buf[i * nj + j] = f(mindex(index.i + i, index.j + j), as.local_get(mindex(i, j)));
      }
    }

    return ret;
  }
    
  //------------------------------------------------------------------------
  // zip and zipwith
  template <typename A, typename B>
  dist_matrix<std::pair<A, B> >
  matrix_skeletons::zip(const dist_matrix<A>& as,
		      const dist_matrix<B>& bs)
  {
    return _impl::ms_adpt::zipwith<sketo::functions::mkpair<A, B>, A, B, std::pair<A, B> >(sketo::functions::mkpair<A, B>(), as, bs);
  }

  template <typename F, typename A, typename B>
  dist_matrix<typename result_of<F(A, B)>::type>
  matrix_skeletons::zipwith(const F& f,
			  const dist_matrix<A>& as,
			  const dist_matrix<B>& bs)
  {
    return _impl::ms_adpt::zipwith<F, A, B, typename result_of<F(A, B)>::type>(f, as, bs);
  }

  template <typename F, typename A, typename B, typename C>
  dist_matrix<C>
  _impl::ms_adpt::zipwith(const F& f,
			  const dist_matrix<A>& as,
			  const dist_matrix<B>& bs)
  {
    assert(as.get_global_size() == bs.get_global_size());
    dist_matrix<C> ret(as.get_global_size());

    C* const ret_buf = ret.data->buf;
    const int nj = as.get_local_size().j;
    const int ni = as.get_local_size().i;
    for (int i = 0; i < ni; ++i) {
      for (int j = 0; j < nj; ++j) {
        ret_buf[i * nj + j] = f(as.local_get(mindex(i, j)), bs.local_get(mindex(i, j)));
      }
    }
    
    return ret;
  }

  template <typename OPLUS, typename OTIMES, typename A>
  dist_matrix<A>
  matrix_skeletons::scan(const OPLUS &oplus,
                         const OTIMES &otimes,
                         const dist_matrix<A>& as)
  {
    dist_matrix<A> bs(as.get_global_size());     
    return _impl::ms_adpt::scan<OPLUS, OTIMES, A>(&bs, oplus, otimes, as);
  }

  template <typename OPLUS, typename OTIMES, typename A>
  dist_matrix<A>
  matrix_skeletons::scanr(const OPLUS &oplus,
                          const OTIMES &otimes,
                          const dist_matrix<A>& as)
  {
    dist_matrix<A> bs(as.get_global_size());         
    return _impl::ms_adpt::scanr<OPLUS, OTIMES, A>(&bs, oplus, otimes, as);
  }

  template <int di, int dj, typename F, typename A>
  dist_matrix<A>
  matrix_skeletons::shift(const F &f,
                          const dist_matrix<A>& as)
  {
    dist_matrix<A> bs(as.get_global_size());
    std::cout << "creating shift_info" << std::endl;
    _impl::shift_info<di, dj, A> si(as.get_global_size());
    std::cout << "shift_info created properly" << std::endl;
    _impl::ms_adpt::shift_common<di, dj, F, dist_matrix<A>, A>(&bs, si, f, as);
    std::cout << "shift_common returned." << std::endl;
    return bs;
  }
 
  
}
  
#endif // __MATRIX_SKELETONS_NOOPT_TPP__
