/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file nqueen_concatmap_filter.cpp
 * @brief Parallel breadth-first-search by concatmap/filter
 */


#include <sketo/list_skeletons.h>
#include <cstdlib>

//-------------- for measuring computation time
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
double gettimeofday_sec()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}
//---------------

struct work {
  unsigned int n;
  unsigned int i;
  unsigned int u;
  unsigned int l;
  unsigned int r;
  unsigned char qs[32];
  work(unsigned int n, unsigned int i, unsigned int u, unsigned int l, unsigned int r, unsigned char *ps) : n(n), i(i), u(u), l(l), r(r) {
    for(int i = 0; i < 32; i++) qs[i] = *ps++;
  }
  work(){}
};

struct step_t {
  template <typename COLLECTOR>
  void operator()(const work &w, COLLECTOR& c) const {
    const unsigned int nl = (w.l << 1) | 1;
    const unsigned int nr = ((w.r|(1<<w.n)) >> 1);
    const unsigned int ni = w.i + 1;
    unsigned int p = w.u & nl & nr;
    while(p!=0) {
      const unsigned int lb = (-p)&p;
      p ^= lb;
      c.push_back(work(w.n, ni, w.u ^ lb, nl ^ lb, nr ^ lb)); 
    }
  }
  typedef work result_type;
  typedef work argument_type;
  typedef work arg1_type;
} step;

struct nextboards_t {
  template <typename COLLECTOR>
  void operator()(const work &w, COLLECTOR& c) const {
    const unsigned int nl = (w.l << 1) | 1;
    const unsigned int nr = ((w.r|(1<<w.n)) >> 1);
    const unsigned int ni = w.i + 1;
    unsigned int p = (1<<w.n);
    unsigned char qs[32];
    for(int i = 0; i < w.i; i++) qs[i] = w.qs[i];
    for(int i = 0; i < w.n; i++) {
      qs[w.i] = i;
      const unsigned int lb = 1 << i;
      c.push_back(work(w.n, ni, w.u ^ lb, nl ^ lb, nr ^ lb, qs)); 
    }
  }
  typedef work result_type;
  typedef work argument_type;
  typedef work arg1_type;
} nextboards;

struct validboard_t {
  bool operator()(const work &w) const {
    const unsigned int lb = 1 << w.qs[w.i-1];
    const unsigned int p = ~(w.u | w.l | w.r) & lb;
    return p;
  }
  typedef work result_type;
  typedef work argument_type;
  typedef work arg1_type;
} validboard;

int sketo::main(int argc, char **argv)
{
  if (argc < 2) {
    sketo::cout << "usage: " << argv[0] << " N [startup_n]" << std::endl;
    return 1;
  }

  const int n = std::atoi(argv[1]); // board size ( i.e. N )
  const int sn = argc > 2 ? std::atoi(argv[2]) : 3; // startup #queens
  
  unsigned int b = (1<<n) - 1;
  unsigned char qs[32];
  const work w(n, 0, b, b, b, qs);   // the empty board

  double st = gettimeofday_sec();
  dist_list<work> bs(1, &w); // the initial list has only the empty board
  int i = 0;
  for(; i < sn; i++) {
    // generate the next valid boards
    bs = 
      sketo::list_skeletons::filter
      (validboard, 
       sketo::list_skeletons::concatmap(nextboards, bs));
  }
  bs.rebalance();
  double st_time = gettimeofday_sec() - st;
  for(; i < n; i++) {
    // generate the next valid boards
    bs = 
      sketo::list_skeletons::filter
      (validboard, 
       sketo::list_skeletons::concatmap(nextboards, bs));
  }
  unsigned long long total = 0;
  total = bs.get_global_size();
  sketo::cout << "SIZE: " << n << "\tTOTAL: " << total << std::endl;
  sketo::cout << "time: " << gettimeofday_sec() - st << " ( " << st_time << " )" << std::endl;

  return 0;
}

/**
sketocxx -O3 nqueen_concatmap_filter.cpp -o nqueen_concatmap_filter_NF -D__SKETO_NO_FUSION__
sketocxx -O3 nqueen_concatmap_filter.cpp -o nqueen_concatmap_filter

./nqueen_concatmap_filter_NF 13 3
SIZE: 13	TOTAL: 73712
time: 9.40605 ( 0.000349045 )

./nqueen_concatmap_filter 13 3
SIZE: 13	TOTAL: 73712
time: 1.50345 ( 0.000241041 )

mpirun -np 2 ./nqueen_concatmap_filter 13 3
SIZE: 13	TOTAL: 73712
time: 0.755715 ( 0.000303984 )


for p in 1 2 4 8; do cmd="mpirun -np $p ./nqueen_concatmap_filter 14 5"; echo $cmd; $cmd; done

mpirun -np 1 ./nqueen_concatmap_filter 14 5
SIZE: 14	TOTAL: 365596
time: 11.0387 ( 0.0120261 )

mpirun -np 2 ./nqueen_concatmap_filter 14 5
SIZE: 14	TOTAL: 365596
time: 5.86178 ( 0.0120821 )

mpirun -np 4 ./nqueen_concatmap_filter 14 5
SIZE: 14	TOTAL: 365596
time: 3.52972 ( 0.0123429 )

mpirun -np 8 ./nqueen_concatmap_filter 14 5
SIZE: 14	TOTAL: 365596
time: 1.96251 ( 0.0134521 )


Probably, we need a wrok-stealing mechanism for good scalability with more processors.

 */
