/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/


/**
 * @file peaksum.cpp
 * @brief Sums up all peaks (elements bigger than their immediate neighbors).
 */

#include <sketo/list_skeletons.h>
#include <cstdlib>


//-------------- for measuring computation time
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
double gettimeofday_sec()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}
//---------------

#define even(a) (!(a&1))

class peak : public sketo::functions::base <bool (std::pair<unsigned int, std::pair<unsigned int, unsigned int> >)> {
public:
  bool operator()(const std::pair<unsigned int, std::pair<unsigned int, unsigned int> > &x) const {
    return (x.first < x.second.first) && (x.second.first > x.second.second);
  }
};

struct rand_gen : sketo::functions::base<unsigned int(int)> {
  rand_gen() {
    //srand((unsigned int)time(NULL));
    srand(0);
  }
  unsigned int operator()(int ) const { return rand(); }
};


int sketo::main(int argc, char **argv)
{
  if (argc < 2) {
    sketo::cout << "usage: " << argv[0] << " length" << std::endl;
    return 1;
  }

  const int n = std::atoi(argv[1]);

  dist_list<unsigned int> x = 
    sketo::list_skeletons::generate(n, rand_gen());
  
  double st = gettimeofday_sec();
  unsigned int r = 
    sketo::list_skeletons::reduce
    (std::plus<unsigned int>(),
     sketo::list_skeletons::map
     (sketo::functions::fst<unsigned int, unsigned int>(),
      sketo::list_skeletons::map
      (sketo::functions::snd<unsigned int, std::pair<unsigned int, unsigned int> >(),
       sketo::list_skeletons::filter
       (peak(),
       sketo::list_skeletons::zip
        (sketo::list_skeletons::shiftr(0U, x),
         sketo::list_skeletons::zip
         (x, sketo::list_skeletons::shiftl(0U, x)))))));
     
  sketo::cout << "SIZE: " << n << "\tpeaksum: " << r << std::endl;
  sketo::cout << "time: " << gettimeofday_sec() - st << std::endl;
  return 0;
}

/**
sketocxx -O3 peaksum.cpp -o peaksum_NF -D__SKETO_NO_FUSION__
sketocxx -O3 peaksum.cpp -o peaksum

./peaksum_NF 100000000
SIZE: 100000000	peaksum: 1245668002
time: 9.85395

./peaksum 100000000
SIZE: 100000000	peaksum: 1245668002
time: 0.613523


Observation and discussion.

The fused peaksum.cpp is 5% slower than the fused peaksum_map.cpp .
This is because the compiled code of the latter uses a 'conditional
move' instruction while one of the former does not. The other parts
are the same.

Maybe, compilers in the future will be strong to replace a conditional
branch in peaksum.cpp with a conditional move instruction, so that
compiled codes of these two become the same.

The compiled main loop of peaksum.cpp:
.L2378:
	movq	40(%r8), %rcx
	movq	(%rcx,%rbp), %rsi
	addq	%rax, %rsi
	movl	(%rsi), %ecx
	movl	4(%rsi), %r9d
	movq	40(%rdi), %rsi
	movq	(%rsi,%rbp), %rsi
	cmpl	(%rsi,%rdx), %ecx
	jbe	.L2377
	cmpl	%r9d, %ecx
	jbe	.L2377
	movq	48(%r12), %rsi
	addl	%ecx, (%rsi)
.L2377:
	addq	$8, %rax
	addq	$4, %rdx
	cmpq	%r10, %rax
	jne	.L2378
	movq	112(%rsp), %rsi
	cmpq	120(%rsp), %rsi
	jne	.L2478

The compiled main loop of peaksum_map.cpp:
.L2467:
	cmpl	%r10d, %eax
	movl	$0, %r8d
	cmovbe	%r8d, %eax
.L2341:
	addl	96(%rsp), %eax
	addq	$8, %rdx
	addq	$4, %rcx
	cmpq	%r9, %rdx
	movl	%eax, 96(%rsp)
	je	.L2326
.L2342:
	movq	40(%rdi), %rax
	movq	%rdx, %r8
	addq	(%rax,%rbp), %r8
	movl	(%r8), %eax
	movl	4(%r8), %r10d
	movq	40(%rsi), %r8
	movq	(%r8,%rbp), %r8
	cmpl	(%r8,%rcx), %eax
	ja	.L2467
	xorl	%eax, %eax
	jmp	.L2341



 */
