/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file test_list_skeletons.cpp
 * @brief Test code for list_skeletons.h and list_skeletons.cpp
 */

#include <mpi.h>
//#define __SKETO_NO_FUSION__

#include "../include/sketo/sketo.h"
#include "../include/sketo/list_skeletons.h"

#include <cassert>
#include <utility>
#include <iostream>
#include <functional>
#include <vector>
#include <string>
#include <cstdio>

const int size = 13;

typedef std::string string;
typedef std::pair<int, int> int_pair;
typedef std::pair<double, double> double_pair;
typedef std::vector<int> int_vector;
typedef std::vector<int_vector> int_mat;

template <typename A, typename B>
std::ostream& operator<<(std::ostream& os, const std::pair<A, B>& x) {
  os << "(" << x.first << "," << x.second << ")";
  return os;
}
template <typename A>
std::ostream& operator<<(std::ostream& os, const std::vector<A>& x) {
  os << "[";
  for(unsigned int i = 0; i < x.size(); i++) {
    if(i > 0) os << ",";
    os << x[i];
  }
  os << "]";
  return os;
}


template <typename T>
int same_arrays(T* as, T* bs, int size)
{
  for (int i = 0; i < size; ++i) {
    if (as[i] != bs[i]) {
      std::cerr << "Wrong data: " << "LHS = " << as[i] << ", RHS = " << bs[i] << ". at index " << i << "." << std::endl;
      return false;
    }
  }
  return true;
}


class to_int_pair_t : public sketo::functions::base <int_pair(int)> {
public: int_pair operator()(int x) const { return int_pair(x + 1, x * x); }
} to_int_pair;

class to_int_vector_t : public sketo::functions::base <int_vector(int)> {
public:
  int_vector operator()(int ii) const {
    int_vector v;
    for(int i = 0; i <= ii; i++) {
      v.push_back(i);
    }
    return v;
  }
} to_int_vector;

class to_string_t : public sketo::functions::base <string(int)> {
public:
  string operator()(int ii) const {
    char buf[25];
    sprintf(buf, "%d", ii);
    string s(buf);
    return s;
  }
} to_string;

// function object for map_index
class vector_append_t : public sketo::functions::base <int_vector(int_vector, int_vector)> {
 public: int_vector operator()(const int_vector &x, const int_vector &y) const {
   int_vector im;
   for(unsigned int i = 0; i < x.size(); i++) im.push_back(x[i]);
   for(unsigned int i = 0; i < y.size(); i++) im.push_back(y[i]);  
   return im;
 }
} vector_append;

int_vector identity_element(const vector_append_t &)
{
  int_vector iv;
  return iv;
}

// function object for map
class same_last_two_bit_t : public sketo::functions::base <int(int_pair)> {
 public: int operator()(int_pair x) const { return ((x.first & 3) == (x.second & 3)); }
} same_last_two_bit;

// function object for map_index
class same_last_two_bit_with_index_t : public sketo::functions::base <int(int, int_pair)> {
 public: int operator()(int x, int_pair y) const { return (((x + y.first) & 3) == (y.second & 3)); }
} same_last_two_bit_with_index;

// function objects for zipwith
class to_double_pair_t : public sketo::functions::base <double_pair(int)> {
 public: double_pair operator()(int x) const { return double_pair(x + 1, x - 1); }
} to_double_pair;

class int_pair_with_doule_pair_t : public sketo::functions::base <int(int_pair, double_pair)> {
 public: int operator()(int_pair x, double_pair y) const { return ((double)x.first == y.first); };
} int_pair_with_double_pair;

// function objcets for gscanl and gscanr
class gscan_f_t : public sketo::functions::base <double (double, int)> {
 public: double operator()(double c, int x) const { return c - x * 2.0; }
} gscan_f;

class gscan_g_t : public sketo::functions::base <double (int)> {
 public: double operator()(int x) const { return x * 2.0; }
} gscan_g;

// function objects for accumulate
class accumulate_g_t : public sketo::functions::base <double_pair (double)> {
 public: double_pair operator()(double x) const { return double_pair(x, x); }
} accumulate_g;

class accumulate_p_t : public sketo::functions::base <double_pair (int, double)> {
public: double_pair operator()(int x, double y) const { return double_pair(-2 * x, y); }
} accumulate_p;

class accumulate_q_t : public sketo::functions::base <double (int)> {
 public: double operator()(int x) const { return -x; };
} accumulate_q;

class accumulate_oplus_t : public sketo::functions::base <double_pair (double_pair, double_pair)> {
 public: double_pair operator()(const double_pair &x, const double_pair &y) const {
   return double_pair(x.first + y.first, x.second + y.second);
 }
} accumulate_oplus;

class first_is_multiple_of : public sketo::functions::base <bool (int_pair)> {
public:
  const int m;
  first_is_multiple_of(int m) : m(m) { }
  bool operator()(const int_pair &x) const {
    return (x.first % m) == 0;
  }
};

class is_multiple_of : public sketo::functions::base <bool (int)> {
public:
  const int m;
  is_multiple_of(int m) : m(m) { }
  bool operator()(const int &x) const {
    return (x % m) == 0;
  }
};

class is_pow2_t : public sketo::functions::base <bool (int)> {
public:
  is_pow2_t() { }
  bool operator()(const int &x) const {
    return (x&(x-1)) == 0;
  }
} is_pow2;

struct _f {
  template <typename COLLECTOR>
  void operator()(int a, COLLECTOR &c) const
  {
    for(int i = 0; i < 3; i++) {
      c.push_back(a * 0.5);
    }
  }
  typedef double result_type;
  typedef int argument_type;
  typedef int arg1_type;
} f;

struct _g {
  template <typename COLLECTOR>
  void operator()(int_pair a, COLLECTOR &c) const
  {
    c.push_back(a.first);
    if(a.first%2 == 0) {
      c.push_back(-a.second);
    }
  }
  typedef int result_type;
  typedef int_pair argument_type;
  typedef int_pair arg1_type;
} g;

struct _succs {
  template <typename COLLECTOR>
  void operator()(int a, COLLECTOR &c) const
  {
    c.push_back(a);
    c.push_back(a+1);
    c.push_back(a+3);
  }
  typedef int result_type;
  typedef int argument_type;
  typedef int arg1_type;
} succs;

// function object for map
class sawp_pair_t : public sketo::functions::base <int_pair(int_pair)> {
public: int_pair operator()(int_pair x) const { return int_pair(x.second, x.first); }
} swap_pair;


void test_get_set()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_list<int_pair>* as = new sketo::dist_list<int_pair>(size);

  for (int i = 0; i < size; ++i) {
    as->set(i, to_int_pair(i));
  }

  for (int i = 0; i < size; ++i) {
    int_pair a = as->get(i);
    if (sketo::is_master()) {
      assert(a == to_int_pair(i));
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_get_set:OK" << std::endl;
}
void test_get_set2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_list<int_vector>* as = new sketo::dist_list<int_vector>(size);

  for (int i = 0; i < size; ++i) {
    as->set(i, to_int_vector(i));
  }
  //as->print();
  for (int i = 0; i < size; ++i) {
    int_vector a = as->get(i);
    if (sketo::is_master()) {
      assert(a == to_int_vector(i));
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_get_set2:OK" << std::endl;
}


void test_push_pop1()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_list<int>* as = new sketo::dist_list<int>(0);

  for (int i = 0; i < size; ++i) {
    as->push_front(i);
  }
  for (int i = 0; i < size; ++i) {
    as->push_back(i);
  }
  //as->print();
  for (int i = size - 1; i >= 0; --i) {
    int k = as->pop_front();
    if (sketo::is_master()) {
      assert(i == k);
    }
  }
  for (int i = size - 1; i >= 0; --i) {
    int k = as->pop_back();
    if (sketo::is_master()) {
      assert(i == k);
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_push_pop1:OK" << std::endl;
}

void test_push_pop2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_list<int>* as = new sketo::dist_list<int>(0);

  for (int i = 0; i < size; ++i) {
    as->push_front(i);
  }
  for (int i = 0; i < size; i++) {
    int k = as->pop_back();
    if (sketo::is_master()) {
      assert(i == k);
    }
  }
  for (int i = 0; i < size; ++i) {
    as->push_back(i);
  }
  for (int i = 0; i < size; i++) {
    int k = as->pop_front();
    if (sketo::is_master()) {
      assert(i == k);
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_push_pop2:OK" << std::endl;
}


void test_append1()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_pair* array = new int_pair[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_int_pair(i);
    else                    array[i] = to_int_pair(i+10);
  }
  
  sketo::dist_list<int_pair> as(size, array);
  sketo::dist_list<int_pair> bs(size/2, array);
  sketo::dist_list<int_pair> cs(size - size/2, array + size/2);

  sketo::dist_list<int_pair> ds = sketo::list_skeletons::append(as, sketo::list_skeletons::append(bs, cs));
  int_pair* array2 = new int_pair[size*2];
  ds.gather(array2);
  
  if (sketo::is_master()) assert(same_arrays(array, array2, size));
  if (sketo::is_master()) assert(same_arrays(array, array2+size, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_append1:OK" << std::endl;
}

void test_rebalance1()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_pair* array = new int_pair[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_int_pair(i);
    else                    array[i] = to_int_pair(i+10);
  }
  
  sketo::dist_list<int_pair> as(size, array);
  sketo::dist_list<int_pair> bs(size/2, array);
  sketo::dist_list<int_pair> cs(size - size/2, array + size/2);

  sketo::dist_list<int_pair> ds = sketo::list_skeletons::append(as, sketo::list_skeletons::append(bs, cs));
  int_pair* array2 = new int_pair[size*2];

  ds.rebalance();

  ds.gather(array2);
  
  if (sketo::is_master()) assert(same_arrays(array, array2, size));
  if (sketo::is_master()) assert(same_arrays(array, array2+size, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_rebalance1:OK" << std::endl;
}

void test_distribute_gather1()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_pair* array = new int_pair[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_int_pair(i);
    else                    array[i] = to_int_pair(i+10);
  }
  
  sketo::dist_list<int_pair> as(size, array);
  for (int i = 0; i < size; ++i) {
    assert(as.get(i) == to_int_pair(i));
  }
  
  int_pair* array2 = new int_pair[size];
  as.gather(array2);
  if (sketo::is_master()) assert(same_arrays(array, array2, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather1:OK" << std::endl;
}

void test_distribute_gather2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_vector* array = new int_vector[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_int_vector(i);
    else                    array[i] = to_int_vector(i+10);
  }
  
  sketo::dist_list<int_vector> as(size, array);
  for (int i = 0; i < size; ++i) {
    assert(as.get(i) == to_int_vector(i));
  }
  
  int_vector* array2 = new int_vector[size];
  as.gather(array2);
  if (sketo::is_master()) assert(same_arrays(array, array2, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather2:OK" << std::endl;
}

void test_distribute_gather3()
{
  MPI_Barrier(MPI_COMM_WORLD);
  string* array = new string[size];
  for (int i = 0; i < size; ++i) {
    if (sketo::is_master()) array[i] = to_string(i);
    else                    array[i] = to_string(i+10);
  }
  
  sketo::dist_list<string> as(size, array);
  for (int i = 0; i < size; ++i) {
    assert(as.get(i) == to_string(i));
  }
  
  string* array2 = new string[size];
  as.gather(array2);
  if (sketo::is_master()) assert(same_arrays(array, array2, size));

  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather3:OK" << std::endl;
}


void test_generate()
{
  MPI_Barrier(MPI_COMM_WORLD);
  // skeleton code
  sketo::dist_list<int_pair> as;
  as = sketo::list_skeletons::generate(size, to_int_pair);
  
  int_pair* array_skel = new int_pair[size];
  as.gather(array_skel);

  // sequential code
  int_pair* array_seq = new int_pair[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = to_int_pair(i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_generate:OK" << std::endl;
}


void test_map()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<int> as;
  as = sketo::list_skeletons::map(same_last_two_bit,
				  sketo::list_skeletons::generate(size, to_int_pair));
  int* array_skel = new int[size];
  as.gather(array_skel);
  
  // sequential code
  int* array_seq = new int[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = same_last_two_bit(to_int_pair(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map:OK" << std::endl;
}


void test_map_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<int> as;
  sketo::dist_list<int_pair> bs = sketo::list_skeletons::generate(size, to_int_pair);

  as = sketo::list_skeletons::map(same_last_two_bit, sketo::list_skeletons::append(bs, bs));

  int* array_skel = new int[size * 2];
  as.gather(array_skel);
  
  // sequential code
  int* array_seq = new int[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = same_last_two_bit(to_int_pair(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(same_arrays(array_skel + size, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map_append:OK" << std::endl;
}

void test_map_overwrite()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<double> as;
  as = sketo::list_skeletons::generate(size, sketo::functions::caster<int, double>());
  as = sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), 2), as);
  double* array_skel = new double[size];
  as.gather(array_skel);

  // sequential code
  double *array_seq = new double[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = 2 * static_cast<double>(i);
  }

  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map_overwrite:OK" << std::endl;
}

void test_map_fusion()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<double> as;
  as = sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), -1),
				  sketo::list_skeletons::map(std::bind1st(std::plus<double>(), -1),
							     sketo::list_skeletons::generate(size, sketo::functions::caster<int, double>())));
}


void test_map_with_index()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<int> bs;
  bs = sketo::list_skeletons::map_with_index(same_last_two_bit_with_index,
					     sketo::list_skeletons::generate(size, to_int_pair));
  int* array_skel = new int[size];
  bs.gather(array_skel);
  
  // sequential code
  int* array_seq = new int[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = same_last_two_bit_with_index(i, to_int_pair(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map_with_index:OK" << std::endl;
}


void test_map_with_index_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<int> bs;
  sketo::dist_list<int_pair> as = sketo::list_skeletons::generate(size, to_int_pair);
  bs = sketo::list_skeletons::map_with_index(same_last_two_bit_with_index,
					     sketo::list_skeletons::append(as, as));
  int* array_skel = new int[size * 2];
  bs.gather(array_skel);
  
  // sequential code
  int* array_seq = new int[size * 2];
  for (int i = 0; i < size * 2; ++i) {
    array_seq[i] = same_last_two_bit_with_index(i, to_int_pair(i%size));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size * 2));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map_with_index_append:OK" << std::endl;
}


void test_zip()
{
  MPI_Barrier(MPI_COMM_WORLD);
  // skeleton code
  std::pair<int_pair, double_pair> *array_skel = new std::pair<int_pair, double_pair>[size];
  {
    sketo::dist_list<std::pair <int_pair, double_pair> > cs;
    cs = sketo::list_skeletons::zip(sketo::list_skeletons::generate(size, to_int_pair),
				    sketo::list_skeletons::generate(size, to_double_pair));
    cs.gather(array_skel);
  }

  // sequential code
  std::pair <int_pair, double_pair> *array_seq = new std::pair <int_pair, double_pair>[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i].first = to_int_pair(i);
    array_seq[i].second = to_double_pair(i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zip:OK" << std::endl;
}

void test_zip_append_push_pop()
{
  MPI_Barrier(MPI_COMM_WORLD);
  // skeleton code
  std::pair<int_pair, double_pair> *array_skel = new std::pair<int_pair, double_pair>[size*3];
  {
    sketo::dist_list<int_pair> as = sketo::list_skeletons::generate(size, to_int_pair);
    sketo::dist_list<double_pair> bs = sketo::list_skeletons::generate(size, to_double_pair);

    for(int i = size-1; i >= 0; i--) as.push_front(to_int_pair(i));
    for(int i = 0; i < size; i++) as.push_back(to_int_pair(i));

    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(bs, bs));

    sketo::dist_list<std::pair <int_pair, double_pair> > cs;
    cs = sketo::list_skeletons::zip(as, bs);
    cs.gather(array_skel);
  }

  // sequential code
  std::pair <int_pair, double_pair> *array_seq = new std::pair <int_pair, double_pair>[size*3];
  for (int i = 0; i < size*3; ++i) {
    array_seq[i].first = to_int_pair(i%size);
    array_seq[i].second = to_double_pair(i%size);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size*3));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zip_append_push_pop:OK" << std::endl;
}

void test_zip_append_push_pop2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  // skeleton code
  std::pair<int_pair, double_pair> *array_skel = new std::pair<int_pair, double_pair>[size*3];
  {
    sketo::dist_list<int_pair> as = sketo::list_skeletons::generate(size, to_int_pair);
    sketo::dist_list<double_pair> bs = sketo::list_skeletons::generate(size, to_double_pair);

    for(int i = size-1; i >= 0; i--) as.push_front(to_int_pair(i));
    for(int i = 0; i < 5; i++) as.push_back(to_int_pair(i));

    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(bs, bs));

    sketo::dist_list<std::pair <int_pair, double_pair> > cs;
    cs = sketo::list_skeletons::zip(as, bs);
    cs.gather(array_skel);
  }

  // sequential code
  std::pair <int_pair, double_pair> *array_seq = new std::pair <int_pair, double_pair>[size*3];
  for (int i = 0; i < size*3; ++i) {
    array_seq[i].first = to_int_pair(i%size);
    array_seq[i].second = to_double_pair(i%size);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size*2+5));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zip_append_push_pop2:OK" << std::endl;
}


void test_zipwith()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> cs;
    cs = sketo::list_skeletons::zipwith(int_pair_with_double_pair,
					sketo::list_skeletons::generate(size, to_int_pair),
					sketo::list_skeletons::generate(size, to_double_pair));
    cs.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = int_pair_with_double_pair(to_int_pair(i), to_double_pair(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zipwith:OK" << std::endl;
}


void test_zipwith_ow()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> cs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    cs = sketo::list_skeletons::zipwith(std::plus<int>(),
					cs,
					sketo::list_skeletons::generate(size, sketo::functions::identity<int>()));
    cs.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 0; i < size; ++i) {
    array_seq[i] = i + i ;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zipwith_ow:OK" << std::endl;
}


void test_concatmap()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<double> as;
  as = sketo::list_skeletons::concatmap(f, 
	  sketo::list_skeletons::concatmap(g,
					   sketo::list_skeletons::generate(size, to_int_pair)));
  double* array_skel = new double[as.get_global_size()];
  as.gather(array_skel);
  
  // sequential code
  std::vector<double> v;
  for (int i = 0; i < size; ++i) {
    std::vector<int> u;
    g(to_int_pair(i), u);
    for(int j = 0; j < u.size(); j++) {
      f(u[j], v);
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(*v.begin()), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_concatmap:OK" << std::endl;
}


void test_concatmap_map()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<double_pair> as;
  as =
    sketo::list_skeletons::map(
      to_double_pair,
    sketo::list_skeletons::concatmap(f, 
	  sketo::list_skeletons::concatmap(g,
                                     sketo::list_skeletons::generate(size, to_int_pair))));
  double_pair* array_skel = new double_pair[as.get_global_size()];
  as.gather(array_skel);
  
  // sequential code
  std::vector<double> vv;
  for (int i = 0; i < size; ++i) {
    std::vector<int> u;
    g(to_int_pair(i), u);
    for(int j = 0; j < u.size(); j++) {
      f(u[j], vv);
    }
  }
  std::vector<double_pair> v;
  for(int i = 0; i < vv.size(); i++) {
    v.push_back(to_double_pair(vv[i]));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(*v.begin()), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_concatmap_map:OK" << std::endl;
}



void test_filter()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<int_pair> as;
  as = sketo::list_skeletons::filter(first_is_multiple_of(3), 
				     sketo::list_skeletons::generate(size, to_int_pair));
  int_pair* array_skel = new int_pair[as.get_global_size()];
  as.gather(array_skel);
  
  // sequential code
  std::vector<int_pair> v;
  for (int i = 0; i < size; ++i) {
    if(first_is_multiple_of(3)(to_int_pair(i))) {
      v.push_back(to_int_pair(i));
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(*v.begin()), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_filter:OK" << std::endl;
}


void test_reduce()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int result_skel = sketo::list_skeletons::reduce(std::plus<int>(),
						  0,
						  sketo::list_skeletons::generate(size, sketo::functions::identity<int>()));

  // sequential code
  int result_seq = 0;
  for (int i = 1; i < size; ++i) { result_seq += i; }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce:OK" << std::endl;
}


void test_reduce_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());

  sketo::dist_list<int> bs(0);
  for(int i = 0; i < size; i++) {
    bs.push_back(i);
  }
  int result_skel = sketo::list_skeletons::reduce(std::plus<int>(),
						  0,
						  sketo::list_skeletons::append(as, bs));

  // sequential code
  int result_seq = 0;
  for (int i = 1; i < size; ++i) { result_seq += i; }

  result_seq *= 2;

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce_append:OK" << std::endl;
}


void test_scan()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int last_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::scan(std::plus<int>(),
							   0,
							   3,
							   sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
							   &last_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int last_seq = 3;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = last_seq; last_seq += i;
  }

  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan:OK" << std::endl;
}


void test_reduce2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector result_skel = sketo::list_skeletons::reduce(vector_append,
							 std::vector<int>(),
                                                         sketo::list_skeletons::generate(size, to_int_vector));

  // sequential code
  int_vector result_seq;
  for (int i = 0; i < size; ++i) { result_seq = vector_append(result_seq, to_int_vector(i)); }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce2:OK" << std::endl;
}


void test_scan2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  int_vector last_skel;
  int_vector e;
  e.push_back(3);
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::scan(vector_append,
								  std::vector<int>(),
							   e,
							   sketo::list_skeletons::generate(size, to_int_vector),
							   &last_skel);
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  int_vector last_seq = e;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = last_seq; last_seq = vector_append(last_seq, to_int_vector(i));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan2:OK" << std::endl;
}


void test_scan_map()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int last_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::map
      (std::bind1st(std::plus<int>(), 2),
       sketo::list_skeletons::scan(std::plus<int>(),
				   0,
				   3,
				   sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
				   &last_skel));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int last_seq = 3;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = last_seq + 2; last_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan_map:OK" << std::endl;
}


void test_scan_map2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_pair* array_skel = new int_pair[size];
  int last_skel;
  {
    sketo::dist_list<int_pair> as =
      sketo::list_skeletons::map
      (to_int_pair, 
       sketo::list_skeletons::map
      (std::bind1st(std::plus<int>(), 2),
       sketo::list_skeletons::scan(std::plus<int>(),
                                   0,
                                   3,
                                   sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
                                   &last_skel)));
       as.gather(array_skel);
  }

  // sequential code
  int_pair *array_seq = new int_pair[size];
  int last_seq = 3;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = to_int_pair(last_seq + 2); last_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan_map2:OK" << std::endl;
}


void test_scan_ow()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int last_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    as = sketo::list_skeletons::scan(std::plus<int>(),
				     0,
				     3,
				     as,
				     &last_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int last_seq = 3;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = last_seq; last_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan_ow:OK" << std::endl;
}


void test_scan_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size*3];
  int last_skel;
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(cs, bs));
    sketo::dist_list<int> as = sketo::list_skeletons::scan(std::plus<int>(),
							   0,
							   3,
							   bs,
							   &last_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size*3];
  int last_seq = 3;
  for (int i = 0; i < size*3; ++i) {
    array_seq[i] = last_seq; last_seq += (i % size);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size*3));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan_append:OK" << std::endl;
}


void test_postscan()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::postscan(std::plus<int>(),
							       0,
							       sketo::list_skeletons::generate(size, sketo::functions::identity<int>()));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  array_seq[0] = 0;
  for (int i = 1; i < size; ++i) {
    array_seq[i] = array_seq[i-1] + i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_postscan:OK" << std::endl;
}


void test_postscan_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size*3];
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(cs, bs));
    sketo::dist_list<int> as = sketo::list_skeletons::postscan(std::plus<int>(),
							       0,
							       bs);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size*3];
  array_seq[0] = 0;
  for (int i = 1; i < size*3; ++i) {
    array_seq[i] = array_seq[i-1] + (i % size);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size*3));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_postscan_append:OK" << std::endl;
}


void test_scanr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int first_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::scanr(std::plus<int>(),
							    0,
							    3,
							    sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
							    &first_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int first_seq = 3;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = first_seq; first_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr:OK" << std::endl;
}


void test_scanr2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  int_vector first_skel;
  int_vector e;
  e.push_back(3);
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::scanr(vector_append,
								   std::vector<int>(),
							    e,
							    sketo::list_skeletons::generate(size, to_int_vector),
							    &first_skel);
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  int_vector first_seq = e;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = first_seq; first_seq = vector_append(to_int_vector(i), first_seq);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr2:OK" << std::endl;
}


void test_scanr_map()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int first_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::map
      (std::bind1st(std::plus<int>(), 2),
       sketo::list_skeletons::scanr(std::plus<int>(),
							    0,
							    3,
							    sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
				    &first_skel));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int first_seq = 3;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = first_seq + 2; first_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr_map:OK" << std::endl;
}


void test_scanr_map2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_pair* array_skel = new int_pair[size];
  int first_skel;
  {
    sketo::dist_list<int_pair> as =
      sketo::list_skeletons::map
      (to_int_pair, 
      sketo::list_skeletons::map
      (std::bind1st(std::plus<int>(), 2),
       sketo::list_skeletons::scanr(std::plus<int>(),
							    0,
							    3,
							    sketo::list_skeletons::generate(size, sketo::functions::identity<int>()),
                                    &first_skel)));
    as.gather(array_skel);
  }

  // sequential code
  int_pair *array_seq = new int_pair[size];
  int first_seq = 3;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = to_int_pair(first_seq + 2); first_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr_map2:OK" << std::endl;
}


void test_scanr_ow()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  int first_skel;
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    as = sketo::list_skeletons::scanr(std::plus<int>(),
				      0,
				      3,
				      as,
				      &first_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  int first_seq = 3;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = first_seq; first_seq += i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr_ow:OK" << std::endl;
}


void test_scanr_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size*size];
  int first_skel;
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(cs, bs));
    sketo::dist_list<int> as = sketo::list_skeletons::scanr(std::plus<int>(),
							    0,
							    3,
							    bs,
							    &first_skel);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size*3];
  int first_seq = 3;
  for (int i = size*3 - 1; i >= 0; --i) {
    array_seq[i] = first_seq; first_seq += (i%size);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size*3));
  if (sketo::is_master()) assert(first_skel == first_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr_append:OK" << std::endl;
}


void test_scan_scanr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  int_vector last_skel;
  int_vector e;
  e.push_back(3);
  {
    sketo::dist_list<int_vector> as =
      sketo::list_skeletons::scanr(vector_append,
                                   e,
                                   sketo::list_skeletons::scan(vector_append,
                                                               std::vector<int>(),
                                                               e,
                                                               sketo::list_skeletons::generate(size, to_int_vector),
                                                               &last_skel));
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  int_vector last_seq = e;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = last_seq; last_seq = vector_append(last_seq, to_int_vector(i));
  }
  int_vector first_seq = e, tmp;
  for (int i = size - 1; i >= 0; i--) {
    tmp = first_seq;
    first_seq = vector_append(array_seq[i], first_seq);
    array_seq[i] = tmp;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  if (sketo::is_master()) assert(last_skel == last_seq);

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan_scanr:OK" << std::endl;
}


void test_postscanr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::postscanr(std::plus<int>(),
								0,
								sketo::list_skeletons::generate(size, sketo::functions::identity<int>()));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  array_seq[size - 1] = size - 1;
  for (int i = size - 1 - 1; i >= 0; --i) {
    array_seq[i] = array_seq[i+1] + i;
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_postscanr:OK" << std::endl;
}


void test_postscanr_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size * 3];
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(cs, bs));
    sketo::dist_list<int> as = sketo::list_skeletons::postscanr(std::plus<int>(),
								0,
								bs);
    as.gather(array_skel);
  }


  // sequential code
  int *array_seq = new int[size * 3];
  array_seq[size * 3 - 1] = size - 1;
  for (int i = size * 3 - 1 - 1; i >= 0; --i) {
    array_seq[i] = array_seq[i+1] + (i % size);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size * 3));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_postscanr_append:OK" << std::endl;
}


void test_gscanl()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  double *array_skel = new double[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<double> bs = sketo::list_skeletons::gscanl(gscan_f, 100.0, gscan_g, std::minus<double>(), std::plus<double>(), 0.0, as);
    bs.gather(array_skel);
  }

  // sequential code
  double *array_seq = new double[size];
  double val = 100;
  for (int i = 0; i < size; ++i) {
    array_seq[i] = val;
    val = gscan_f(val, i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_gscanl:OK" << std::endl;
}    


void test_gscanl_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  double *array_skel = new double[size * 3];
  {

    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    as = sketo::list_skeletons::append(as, sketo::list_skeletons::append(cs, as));

    sketo::dist_list<double> bs = sketo::list_skeletons::gscanl(gscan_f, 100.0, gscan_g, std::minus<double>(), std::plus<double>(), 0.0, as);
    bs.gather(array_skel);
  }

  // sequential code
  double *array_seq = new double[size * 3];
  double val = 100;
  for (int i = 0; i < size * 3; ++i) {
    array_seq[i] = val;
    val = gscan_f(val, (i % size));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size * 3));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_gscanl_append:OK" << std::endl;
}    
		       

void test_gscanr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  double *array_skel = new double[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<double> bs = sketo::list_skeletons::gscanr(gscan_f, 100.0, gscan_g, std::minus<double>(), std::plus<double>(), 0.0, as);
    bs.gather(array_skel);
  }

  // sequential code
  double *array_seq = new double[size];
  double val = 100;
  for (int i = size - 1; i >= 0; --i) {
    array_seq[i] = val;
    val = gscan_f(val, i);
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_gscanr:OK" << std::endl;
}    


void test_gscanr_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  double *array_skel = new double[size*3];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    as = sketo::list_skeletons::append(as, sketo::list_skeletons::append(cs, as));

    sketo::dist_list<double> bs = sketo::list_skeletons::gscanr(gscan_f, 100.0, gscan_g, std::minus<double>(), std::plus<double>(), 0.0, as);
    bs.gather(array_skel);
  }

  // sequential code
  double *array_seq = new double[size*3];
  double val = 100;
  for (int i = size*3 - 1; i >= 0; --i) {
    array_seq[i] = val;
    val = gscan_f(val, (i%size));
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size*3));

  delete [] array_skel;
  delete [] array_seq;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_gscanr_append:OK" << std::endl;
}    
		       

void test_shiftl()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::shiftl(11,
				 sketo::list_skeletons::shiftl(10,
				   sketo::list_skeletons::generate(size, sketo::functions::identity<int>())));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 0; i < size - 2; ++i) {
    array_seq[i] = i + 2;
  }
  array_seq[size - 2] = 10;
  array_seq[size - 1] = 11;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl:OK" << std::endl;
}


void test_shiftr()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::shiftr(11,
				 sketo::list_skeletons::shiftr(10,
				   sketo::list_skeletons::generate(size, sketo::functions::identity<int>())));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 2; i < size; ++i) {
    array_seq[i] = i - 2;
  }
  array_seq[0] = 11;
  array_seq[1] = 10;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr:OK" << std::endl;
}


void test_shiftl2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::shiftl(to_int_vector(11),
                                                                    sketo::list_skeletons::shiftl(to_int_vector(10),
				   sketo::list_skeletons::generate(size, to_int_vector)));
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  for (int i = 0; i < size - 2; ++i) {
    array_seq[i] = to_int_vector(i + 2);
  }
  array_seq[size - 2] = to_int_vector(10);
  array_seq[size - 1] = to_int_vector(11);
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));
  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl2:OK" << std::endl;
}


void test_shiftr2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_vector* array_skel = new int_vector[size];
  {
    sketo::dist_list<int_vector> as = sketo::list_skeletons::shiftr(to_int_vector(11),
                                                                    sketo::list_skeletons::shiftr(to_int_vector(10),
				   sketo::list_skeletons::generate(size, to_int_vector)));
    as.gather(array_skel);
  }

  // sequential code
  int_vector *array_seq = new int_vector[size];
  for (int i = 2; i < size; ++i) {
    array_seq[i] = to_int_vector(i - 2);
  }
  array_seq[0] = to_int_vector(11);
  array_seq[1] = to_int_vector(10);
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr2:OK" << std::endl;
}


void test_shiftl_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size*3];
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(cs, bs));

    sketo::dist_list<int> as = sketo::list_skeletons::shiftl(11,
				 sketo::list_skeletons::shiftl(10,
				   bs));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size*3];
  for (int i = 0; i < size*3 - 2; ++i) {
    array_seq[i] = ((i + 2) % size);
  }
  array_seq[size*3 - 2] = 10;
  array_seq[size*3 - 1] = 11;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size*3));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl_append:OK" << std::endl;
}


void test_shiftr_append()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size*3];
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    sketo::dist_list<int> cs(0);
    for(int i = 0 ; i < size; i++) {
      cs.push_back(i);
    }
    
    bs = sketo::list_skeletons::append(bs, sketo::list_skeletons::append(cs, bs));
    sketo::dist_list<int> as = sketo::list_skeletons::shiftr(11,
				 sketo::list_skeletons::shiftr(10,
				   bs));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size*3];
  for (int i = 2; i < size*3; ++i) {
    array_seq[i] = ((i-2) % size);
  }
  array_seq[0] = 11;
  array_seq[1] = 10;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size * 3));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr_append:OK" << std::endl;
}


void test_shiftl_overwrite()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    as = sketo::list_skeletons::shiftl(10,
				       as);
    as = sketo::list_skeletons::shiftl(11,
				       as);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 0; i < size - 2; ++i) {
    array_seq[i] = i + 2;
  }
  array_seq[size - 2] = 10;
  array_seq[size - 1] = 11;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl_overwrite:OK" << std::endl;
}


void test_shiftr_overwrite()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> as = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    as = sketo::list_skeletons::shiftr(10,
				       as);
    as = sketo::list_skeletons::shiftr(11,
				       as);
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size];
  for (int i = 2; i < size; ++i) {
    array_seq[i] = i - 2;
  }
  array_seq[0] = 11;
  array_seq[1] = 10;
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr_overwrite:OK" << std::endl;
}


void test_shiftl_filter()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size*3];
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    bs = sketo::list_skeletons::filter(is_multiple_of(3), bs);
    sketo::dist_list<int> as = sketo::list_skeletons::shiftl(11,
				 sketo::list_skeletons::shiftl(10,
				   bs));
    as.gather(array_skel);
  }

  // sequential code
  std::vector<int> v;
  int k = 0;
  for (int i = 0; i < size; ++i) {
    if(is_multiple_of(3)(i)) if(++k>2)v.push_back(i);
  }
  v.push_back(10);
  v.push_back(11);
  
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(v[0]), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl_filter:OK" << std::endl;
}


void test_shiftr_filter()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int* array_skel = new int[size];
  {
    sketo::dist_list<int> bs = sketo::list_skeletons::generate(size, sketo::functions::identity<int>());
    bs = sketo::list_skeletons::filter(is_multiple_of(3), bs);
    sketo::dist_list<int> as = sketo::list_skeletons::shiftr(11,
				 sketo::list_skeletons::shiftr(10,
				   bs));
    as.gather(array_skel);
  }

  // sequential code
  std::vector<int> v;
  v.push_back(11);
  v.push_back(10);
  for (int i = 0; i < size; ++i) {
    if(is_multiple_of(3)(i)) v.push_back(i);
  }

  v.pop_back();
  v.pop_back();
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(v[0]), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr_filter:OK" << std::endl;
}



void test_shiftr_map_concatmap()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  
  sketo::dist_list<int> bs = 
    sketo::list_skeletons::shiftr
    (10,
     sketo::list_skeletons::map
     (std::bind1st(std::plus<int>(), 2),
      sketo::list_skeletons::map
      (std::bind1st(std::plus<int>(), 2),
       sketo::list_skeletons::concatmap
       (succs, 
        sketo::list_skeletons::concatmap
        (succs, 
         sketo::list_skeletons::generate(size, sketo::functions::identity<int>()))))));

  int* array_skel = new int[bs.get_global_size()];
  bs.gather(array_skel);
  
  std::vector<int> vvv;
  for (int i = 0; i < size; ++i) {
    succs(i, vvv);
  }

  std::vector<int> vv;
  for (int i = 0; i < vvv.size(); ++i) {
    succs(vvv[i], vv);
  }

  std::vector<int> v;
  v.push_back(10);
  for (int i = 0; i < vv.size()-1; ++i) {
    v.push_back(vv[i] + 2 + 2);
  }
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(v[0]), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftr_map_concatmap:OK" << std::endl;
}


void test_zip_concatmap()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  
  sketo::dist_list<int_pair > bs = 
  sketo::list_skeletons::zip
    (sketo::list_skeletons::filter
     (is_multiple_of(3),
      sketo::list_skeletons::generate(size, sketo::functions::identity<int>())),
     sketo::list_skeletons::filter
     (is_pow2,
      sketo::list_skeletons::generate(size, sketo::functions::identity<int>())));
  
  int_pair* array_skel = new int_pair[bs.get_global_size()];
  bs.gather(array_skel);

  // sequential code
  std::vector<int> v1, v2;
  for (int i = 0; i < size; ++i) {
    if(is_multiple_of(3)(i)) v1.push_back(i);
    if(is_pow2(i)) v2.push_back(i);
  }

  std::vector<int_pair> v;
  for (int i = 0; i < v1.size() && i < v2.size(); ++i) {
    v.push_back(int_pair(v1[i], v2[i]));
  }

   // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(v[0]), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zip_concatmap:OK" << std::endl;
}


void test_shiftl_shiftr_zip_map_filter_concatmap()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  
  sketo::dist_list<int> bs = 
    sketo::list_skeletons::shiftr
    (10,
     sketo::list_skeletons::shiftl
     (-20,
      sketo::list_skeletons::shiftl
      (11,
       sketo::list_skeletons::shiftr
       (-10,
	sketo::list_skeletons::map
	(std::bind1st(std::plus<int>(), 2),
	 sketo::list_skeletons::concatmap
	 (succs, 
	  sketo::list_skeletons::concatmap
	  (g, 
	   (sketo::list_skeletons::map
	    (swap_pair,
	     (sketo::list_skeletons::map
	      (swap_pair,
	       sketo::list_skeletons::zip
	       (sketo::list_skeletons::map
		(std::bind1st(std::multiplies<int>(), 2),
		 (sketo::list_skeletons::map
		  (std::bind1st(std::plus<int>(), -1),
		   sketo::list_skeletons::filter
		   (is_multiple_of(3),
		    sketo::list_skeletons::generate(size, sketo::functions::identity<int>()))))),
		sketo::list_skeletons::filter
		(is_pow2,
		 sketo::list_skeletons::generate(size, sketo::functions::identity<int>()))))))))))))));

  int* array_skel = new int[bs.get_global_size()];
  bs.gather(array_skel);

  // sequential code
  std::vector<int> v1, v2;
  for (int i = 0; i < size; ++i) {
    if(is_multiple_of(3)(i)) v1.push_back(2 * (i-1));
    if(is_pow2(i)) v2.push_back(i);
  }

  std::vector<int> vv;
  for (int i = 0; i < v1.size() && i < v2.size(); ++i) {
    g(int_pair(v1[i], v2[i]), vv);
  }

  std::vector<int> v;
  for (int i = 0; i < vv.size(); ++i) {
    succs(vv[i], v);
  }
  for (int i = 0; i < v.size(); ++i) {
    v[i] += 2;
  }
  v[0] = 10;
  v.back() = 11;
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, &(v[0]), v.size()));

  delete [] array_skel;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shiftl_shiftr_zip_map_filter_concatmap:OK" << std::endl;
}


void test_reduce_map_concatmap_zip_map_filter_concatmap()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code

  int res =
    sketo::list_skeletons::reduce
    (std::plus<int>(), 0,
     sketo::list_skeletons::map
     (std::bind1st(std::plus<int>(), 2),
      sketo::list_skeletons::concatmap
      (succs, 
       sketo::list_skeletons::concatmap
       (g, 
        (sketo::list_skeletons::map
         (swap_pair,
          (sketo::list_skeletons::map
           (swap_pair,
            sketo::list_skeletons::zip
            (sketo::list_skeletons::map
             (std::bind1st(std::multiplies<int>(), 2),
              (sketo::list_skeletons::map
               (std::bind1st(std::plus<int>(), -1),
                sketo::list_skeletons::filter
                (is_multiple_of(3),
                 sketo::list_skeletons::generate(size, sketo::functions::identity<int>()))))),
             sketo::list_skeletons::filter
             (is_pow2,
              sketo::list_skeletons::generate(size, sketo::functions::identity<int>())))))))))));

  // sequential code
  std::vector<int> v1, v2;
  for (int i = 0; i < size; ++i) {
    if(is_multiple_of(3)(i)) v1.push_back(2 * (i-1));
    if(is_pow2(i)) v2.push_back(i);
  }

  std::vector<int> vv;
  for (int i = 0; i < v1.size() && i < v2.size(); ++i) {
    g(int_pair(v1[i], v2[i]), vv);
  }

  std::vector<int> v;
  for (int i = 0; i < vv.size(); ++i) {
    succs(vv[i], v);
  }
  for (int i = 0; i < v.size(); ++i) {
    v[i] += 2;
  }
  int r = 0;
  for(int i = 0; i < v.size(); ++i) {
    r += v[i];
  }
  // compare
  if (sketo::is_master()) assert(r == res);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce_map_concatmap_zip_map_filter_concatmap:OK" << std::endl;
}


// 
// void test_accumulate()
// {
//   MPI_Barrier(MPI_COMM_WORLD);
//   // initial list
//   int *array = new int[size];
//   for (int i = 0; i < size; ++i) { array[i] = i + 1; }
    
//   // skeleton code
//   sketo::dist_list<int> *as = new sketo::dist_list<int>(size, array);

//   double_pair result_skel = sketo::list_skeletons::accumulate(accumulate_g, accumulate_p, accumulate_oplus, accumulate_q, std::multiplies<double>(), 10.0, as);

//   // sequential code
//   double *array_seq = new double[size];
//   array_seq[0] = 10.0;
//   for (int i = 1; i < size; ++i) { array_seq[i] = array_seq[i-1] * (-array[i-1]); }
//   double last = array_seq[size-1] * (-array[size-1]);
//   double_pair val(last, last);
//   for (int i = size - 1; i >= 0; --i) {
//     val.first -= 2 * array[i];
//     val.second += array_seq[i];
//   }
    
//   // compare
// //   std::cout << "LHS = " << result_skel << ", RHS = " << val << std::endl;
//   if (sketo::is_master()) { assert(result_skel == val); }

//   delete [] array_seq;
//   delete [] array;
//   delete as;

//   MPI_Barrier(MPI_COMM_WORLD);
//   sketo::cout << __FILE__ << ":test_accumulate:OK" << std::endl;
// }

#ifdef __GXX_EXPERIMENTAL_CXX0X__  // using c++0x
void test_lambda() {
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int result_skel = sketo::list_skeletons::reduce(
    [](int a, int b) { return a + b;},
    0,
    sketo::list_skeletons::map(
      [](int a) { return a * a;},
      sketo::list_skeletons::generate(size,
                                      [](int a) { return a;})));

  // sequential code
  int result_seq = 0;
  for (int i = 1; i < size; ++i) { result_seq += i*i; }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_lambda:OK" << std::endl;
}

void test_auto_fusion() {
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code

  auto x1 = sketo::list_skeletons::generate(size, [](int a) { return a;});
  auto x2 = sketo::list_skeletons::map([](int a) { return a * a;}, x1);
  int result_skel = sketo::list_skeletons::reduce([](int a, int b) { return a + b;}, 0, x2);

  // sequential code
  int result_seq = 0;
  for (int i = 1; i < size; ++i) { result_seq += i*i; }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_auto_fusion:OK" << std::endl;
}
#endif



int sketo::main(int, char**)
{
  test_get_set();
  test_get_set2();
  test_distribute_gather1();
  test_distribute_gather2();
  test_distribute_gather3();
  
  test_push_pop1();
  test_push_pop2();

  test_append1();
  test_rebalance1();

  test_generate();
  test_map();
  test_map_append();
  test_map_fusion();
  test_map_overwrite();
  test_map_with_index();
  test_map_with_index_append();  // internally, zip is used.
  test_zip();
  test_zip_append_push_pop();
  test_zip_append_push_pop2();
  test_zipwith();
  test_zipwith_ow();

  test_concatmap();
  test_filter();
  test_concatmap_map();
  
  
  test_reduce();
  test_reduce2();
  test_reduce_append();

  test_scan();
  test_scan2();
  test_scan_map();
  test_scan_map2();
  test_scan_ow();
  test_scan_append();

  test_postscan();
  test_postscan_append();


  test_scanr();
  test_scanr2();
  test_scanr_map();
  test_scanr_map2();
  test_scanr_ow();
  test_scanr_append();

  test_scan_scanr();

  test_postscanr();
  test_postscanr_append();

  test_gscanl();
  test_gscanl_append();

  test_gscanr();
  test_gscanr_append();

  test_shiftl();
  test_shiftr();
  test_shiftl2();
  test_shiftr2();
  test_shiftl_append();
  test_shiftr_append();

  test_shiftl_overwrite();
  test_shiftr_overwrite();

  test_shiftl_filter();
  test_shiftr_filter();

  test_shiftr_map_concatmap();

  test_zip_concatmap();

  test_shiftl_shiftr_zip_map_filter_concatmap();
  test_reduce_map_concatmap_zip_map_filter_concatmap();


//   test_accumulate();
  
// to test the lambda expression, we should compile the test program by hand:
// /usr/local/mpich2-1.2.1p1/bin//mpicxx -L../src -lsketo -I../include -O3 test_list_skeletons.cpp -o test_list_skeletons.sketo -std=gnu++0x

#ifdef __GXX_EXPERIMENTAL_CXX0X__  // using c++0x
  test_lambda();
  test_auto_fusion();
#endif

  return 0;
}
